#
# tip_dlg.py <Johannes.Baiter@gmail.com>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.pyqt_ui.ui_tip_dlg import Ui_TipDlg
from mnemosyne.libmnemosyne.ui_components.dialogs import TipDialog


class TipDlg(QtWidgets.QDialog, TipDialog, Ui_TipDlg):

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.tips = []
        self.tips.append(_("""For optimal results, it's best to do your repetitions every day."""))
        self.tips.append(_("""You don't need to finish all your daily scheduled repetitions in a single session."""))
        self.tips.append(_("""If you've been away for a few days, don't worry about your backlog. Do as many cards as you feel like to catch up, the rest will be automatically rescheduled to the future in the most optimal way."""))
        self.tips.append(_("""Sister cards are cards which are based on the same information, e.g. a front-to-back card and the corresponding back-to-front card. Mnemosyne will avoid scheduling these on the same day."""))
        self.tips.append(_("""The 'Work on at most X non-memorised cards at the same time' setting determines how many cards you are trying to (re)learn at the same time. It does <b>not</b> tell you how many new cards you need to learn per day. You are the judge of that: you can learn more cards or less cards, depending on how you feel."""))
        self.tips.append(_("""In summary, try to do your repetitions every day, but don't worry too much about getting the 'Scheduled' counter to zero, and certainly not about getting the 'Not memorised' counter to zero."""))
        self.tips.append(_("""Grade 1 cards are different from grade 0 cards in the sense that they show up less often."""))
        self.tips.append(_("""Use 'Learn ahead of schedule' sparingly. For cramming before an exam, it's much better to use the cramming scheduler plugin."""))
        self.tips.append(_("""You can use keyboard shortcuts to do your repetitions. Enter, Return or Space stand for the default action. The number keys can be used for grading."""))
        self.tips.append(_("""You can select which cards you wish to study in the '(De)activate cards' menu option."""))
        self.tips.append(_("""It is recommended to put all your cards in a single database and use tag to organise them. Using '(De)activate cards' is much more convenient than having to load and unload several databases."""))
        self.tips.append(_("""You can add multiple tags to a card by separating tags with a comma in the 'Tag(s)' input field."""))
        self.tips.append(_("""You can organise tags in a hierarchy by using :: as separator, e.g. My book::Lesson 1."""))
        self.tips.append(_("""You can add images and sounds to your cards. Right-click on an input field when editing a card to bring up a pop-up menu to do so."""))
        self.tips.append(_("""You can make clones of existing card types. This allows you to format cards in this type independently from cards in the original type. E.g. you can make a clone of 'Vocabulary', call it 'Thai' and set a Thai font specifically for this card type without disturbing your other cards."""))
        self.tips.append(_("""If for a certain card type cloned from Vocabulary you don't need a pronunciation field, you can hide it by right-clicking on it and using the pop-up menu."""))
        self.tips.append(_("""You can use basic HTML tags in your cards to control their appearance. However, if you want all the fields in a certain card type to look the same, it's easier to use the 'Set card appearance' menu option."""))
        self.tips.append(_("""Using 'File - Sync', you can sync this machine with a remote server. Of course, that remote computer needs to have a server running, which can be started from the configuration screen on that remote machine."""))
        self.tips.append(_(""" If you want to sync a mobile device with this computer, don't use 'File - Sync', but first enable a sync server in the configuration dialog, and then start the sync from the mobile device."""))
        self.tips.append(_("""In the 'Activate cards' dialog, you can right-click on a saved set to rename or delete it."""))
        self.tips.append(_("""In the 'Activate cards' dialog, you can double-click on a saved set to activate it and close the dialog."""))
        self.tips.append(_("""Right-click on a tag name in the card browser to edit or delete it."""))
        self.tips.append(_("""Double-click on a card or tag name in the card browser to edit them."""))
        self.tips.append(_("""You can reorder columns in the card browser by dragging the header label."""))
        self.tips.append(_("""You can resize columns in the card browser by dragging between the header labels."""))
        self.tips.append(_("""In the card browser, cards with strike-through text are inactive in the current set."""))
        self.tips.append(_("""When editing or previewing cards from the card browser, PageUp/PageDown can be used to move to the previous/next card."""))
        self.tips.append(_("""In the search box of the card browser, you can use SQL wildcards like _ (matching a single character) and % (matching one or more characters)."""))
        self.tips.append(_("""In the 'Add cards' dialog, use Tab to move between different fields, Ctrl+Enter for 'Yet to learn', and Ctrl+2, etc. for the grades."""))
        self.tips.append(_("""In the 'Edit card' dialog, use Tab to move between different fields and Ctrl+Enter to close the dialog and accept the changes."""))
        self.tips.append(_("""Double-click on the name of a saved set in '(De)activate cards' to quickly activate it and close the dialog."""))
        self.tips.append(_("""If you single-click the name of a saved set in '(De)activate cards', modifications to the selected tags and card types are not saved to that set unless you press 'Save this set for later use' again. This allows you to make some quick-and-dirty temporary modifications."""))
        self.tips.append(_("""Mnemosyne can use LaTeX to render mathematical formulas, e.g. <$>x^2+y^2=z^2</$>. (For this, you need LaTeX and dvipng installed.)"""))
        self.tips.append(_("""The best way to backup your data is to copy your mnemosyne data directory and move it to a different drive. Mnemosyne keeps automatic backups, but that won't help you if that drive dies..."""))
        self.tips.append(_("""You can sort the cards in the 'Browse cards' dialog by by clicking on a column title. Clicking again changes the sort order."""))
        self.tips.append(_("""If you want more fine-grained control over LaTeX's behaviour, see the explanation of the <$$>...</$$> and <latex>...</latex> tags on Mnemosyne's website."""))
        self.tips.append(_("""For optimal performance, keep your drives defragmented."""))
        self.tips.append(_("""For optimal performance, do not put your database on a network drive."""))
        self.tips.append(_("""For optimal performance, run 'File - Compact' from time to time, especially after deleting many cards."""))
        self.tips.append(_("""Advanced users can customise more of Mnemosyne by editing the config.py file in their mnemosyne directory. They can also install additional plugins to customise Mnemosyne even further."""))
        self.tips.append(_("""You can follow the development of Mnemosyne on Twitter or Facebook."""))
        self.tips.append(_("""You can request new features and vote for exisiting requests at <a href="https://mnemosyne.uservoice.com/">uservoice</a>. This helps the developers decide what to work on next."""))
        self.tips.append(_("""Using the 'Manage card types' dialog, you can associate a language with a card type. This enables functionality like text-to-speech or dictionary lookup."""))
        self.tips.append(_("""If you've been using Mnemosyne for a long time, and your daily workload becomes too much, consider using the option to stop showing cards after they reach a number of successful repetitions."""))

        self.setupUi(self)
        self.setWindowFlags(self.windowFlags() \
            | QtCore.Qt.WindowMinMaxButtonsHint)
        self.setWindowFlags(self.windowFlags() \
            & ~ QtCore.Qt.WindowContextHelpButtonHint)
        if self.config()["show_daily_tips"] == True:
            self.show_tips.setCheckState(QtCore.Qt.Checked)
        else:
            self.show_tips.setCheckState(QtCore.Qt.Unchecked)
        # Note: the svg file does not seem to work under windows.
        #watermark = QtGui.QPixmap(":/mnemosyne/pixmaps/mnemosyne.svg").\
        #    scaledToHeight(200, QtCore.Qt.SmoothTransformation)
        watermark = QtGui.QPixmap(":/mnemosyne/pixmaps/mnemosyne.png")
        self.watermark.setPixmap(watermark)
        self.update_dialog()

    def activate(self):
        TipDialog.activate(self)
        self.exec_()

    def update_dialog(self):
        # We need an extra modulo operation here to deal with the possibility
        # of decreasing the number of tips during upgrade.
        tip = self.config()["current_tip"] % len(self.tips)
        self.tip_label.setText(self.tips[tip])
        self.previous_button.setEnabled(tip != 0)
        self.next_button.setEnabled(tip != len(self.tips) - 1)

    def previous(self):
        self.config()["current_tip"] = \
            (self.config()["current_tip"] - 1) % len(self.tips)
        self.update_dialog()

    def next(self):
        self.config()["current_tip"] = \
            (self.config()["current_tip"] + 1) % len(self.tips)
        self.update_dialog()

    def reject(self):
        self.close()

    def closeEvent(self, event):
        self.config()["show_daily_tips"] = self.show_tips.isChecked()
        self.config()["current_tip"] = \
            (self.config()["current_tip"] + 1) % len(self.tips)
        event.accept()



import re
from ..utils import stripHTML, stripHTMLMedia
from ..hooks import runFilter
from .furigana import install as furigana_install; furigana_install()
from .hint import install as hint_install; hint_install()

clozeReg = r"(?s)\{\{c%s::(.*?)(::(.*?))?\}\}"

modifiers = {}
def modifier(symbol):
    """Decorator for associating a function with a Mustache tag modifier.

    @modifier('P')
    def render_tongue(self, tag_name=None, context=None):
        return ":P %s" % tag_name

    {{P yo }} => :P yo
    """
    def set_modifier(func):
        modifiers[symbol] = func
        return func
    return set_modifier


def get_or_attr(obj, name, default=None):
    try:
        return obj[name]
    except KeyError:
        return default
    except:
        try:
            return getattr(obj, name)
        except AttributeError:
            return default


class Template:
    # The regular expression used to find a #section
    section_re = None

    # The regular expression used to find a tag.
    tag_re = None

    # Opening tag delimiter
    otag = '{{'

    # Closing tag delimiter
    ctag = '}}'

    def __init__(self, template, context=None):
        self.template = template
        self.context = context or {}
        self.compile_regexps()

    def render(self, template=None, context=None, encoding=None):
        """Turns a Mustache template into something wonderful."""
        template = template or self.template
        context = context or self.context

        template = self.render_sections(template, context)
        result = self.render_tags(template, context)
        if encoding is not None:
            result = result.encode(encoding)
        return result

    def compile_regexps(self):
        """Compiles our section and tag regular expressions."""
        tags = { 'otag': re.escape(self.otag), 'ctag': re.escape(self.ctag) }

        section = r"%(otag)s[\#|^]([^\}]*)%(ctag)s(.+?)%(otag)s/\1%(ctag)s"
        self.section_re = re.compile(section % tags, re.M|re.S)

        tag = r"%(otag)s(#|=|&|!|>|\{)?(.+?)\1?%(ctag)s+"
        self.tag_re = re.compile(tag % tags)

    def render_sections(self, template, context):
        """Expands sections."""
        while 1:
            match = self.section_re.search(template)
            if match is None:
                break

            section, section_name, inner = match.group(0, 1, 2)
            section_name = section_name.strip()

            # check for cloze
            val = None
            m = re.match("c[qa]:(\d+):(.+)", section_name)
            if m:
                # get full field text
                txt = get_or_attr(context, m.group(2), None)
                m = re.search(clozeReg%m.group(1), txt)
                if m:
                    val = m.group(1)
            else:
                val = get_or_attr(context, section_name, None)

            replacer = ''
            inverted = section[2] == "^"
            if (val and not inverted) or (not val and inverted):
                replacer = inner

            template = template.replace(section, replacer)

        return template

    def render_tags(self, template, context):
        """Renders all the tags in a template for a context."""
        while 1:
            match = self.tag_re.search(template)
            if match is None:
                break

            tag, tag_type, tag_name = match.group(0, 1, 2)
            tag_name = tag_name.strip()
            try:
                func = modifiers[tag_type]
                replacement = func(self, tag_name, context)
                template = template.replace(tag, replacement)
            except (SyntaxError, KeyError):
                return "{{invalid template}}"

        return template

    # {{{ functions just like {{ in anki
    @modifier('{')
    def render_tag(self, tag_name, context):
        return self.render_unescaped(tag_name, context)

    @modifier('!')
    def render_comment(self, tag_name=None, context=None):
        """Rendering a comment always returns nothing."""
        return ''

    @modifier(None)
    def render_unescaped(self, tag_name=None, context=None):
        """Render a tag without escaping it."""
        txt = get_or_attr(context, tag_name)
        if txt is not None:
            # some field names could have colons in them
            # avoid interpreting these as field modifiers
            # better would probably be to put some restrictions on field names
            return txt

        # field modifiers
        parts = tag_name.split(':')
        extra = None
        if len(parts) == 1 or parts[0] == '':
            return '{unknown field %s}' % tag_name
        else:
            mods, tag = parts[:-1], parts[-1] #py3k has *mods, tag = parts

        txt = get_or_attr(context, tag)
        
        #Since 'text:' and other mods can affect html on which Anki relies to
        #process clozes, we need to make sure clozes are always
        #treated after all the other mods, regardless of how they're specified
        #in the template, so that {{cloze:text: == {{text:cloze:
        #For type:, we return directly since no other mod than cloze (or other
        #pre-defined mods) can be present and those are treated separately
        mods.reverse()
        mods.sort(key=lambda s: not s=="type")

        for mod in mods:
            # built-in modifiers
            if mod == 'text':
                # strip html
                txt = stripHTML(txt) if txt else ""
            elif mod == 'type':
                # type answer field; convert it to [[type:...]] for the gui code
                # to process
                return "[[%s]]" % tag_name
            elif mod.startswith('cq-') or mod.startswith('ca-'):
                # cloze deletion
                mod, extra = mod.split("-")
                txt = self.clozeText(txt, extra, mod[1]) if txt and extra else ""
            else:
                # hook-based field modifier
                mod, extra = re.search("^(.*?)(?:\((.*)\))?$", mod).groups()
                txt = runFilter('fmod_' + mod, txt or '', extra or '', context,
                                tag, tag_name);
                if txt is None:
                    return '{unknown field %s}' % tag_name
        return txt

    def clozeText(self, txt, ord, type):
        reg = clozeReg
        if not re.search(reg%ord, txt):
            return ""
        def repl(m):
            # replace chosen cloze with type
            if type == "q":
                if m.group(3):
                    return "<span class=cloze>[%s]</span>" % m.group(3)
                else:
                    return "<span class=cloze>[...]</span>"
            else:
                return "<span class=cloze>%s</span>" % m.group(1)
        txt = re.sub(reg%ord, repl, txt)
        # and display other clozes normally
        return re.sub(reg%"\d+", "\\1", txt)

    @modifier('=')
    def render_delimiter(self, tag_name=None, context=None):
        """Changes the Mustache delimiter."""
        try:
            self.otag, self.ctag = tag_name.split(' ')
        except ValueError:
            # invalid
            return
        self.compile_regexps()
        return ''



#
# change_card_type_dlg.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.component import Component
from mnemosyne.pyqt_ui.ui_change_card_type_dlg import Ui_ChangeCardTypeDlg


class ChangeCardTypeDlg(QtWidgets.QDialog, Component, Ui_ChangeCardTypeDlg):

    def __init__(self, current_card_type, return_values, **kwds):
        super().__init__(**kwds)
        self.setupUi(self)
        self.setWindowFlags(self.windowFlags() \
            | QtCore.Qt.WindowMinMaxButtonsHint)
        self.setWindowFlags(self.windowFlags() \
            & ~ QtCore.Qt.WindowContextHelpButtonHint)
        self.card_type_by_name = {}
        self.return_values = return_values
        for card_type in self.database().sorted_card_types():
            if card_type == current_card_type or card_type.hidden_from_UI:
                continue
            self.card_type_by_name[_(card_type.name)] = card_type
            self.card_types_widget.addItem(_(card_type.name))

    def accept(self):
        card_type_name = self.card_types_widget.currentText()
        self.return_values["new_card_type"] = \
            self.card_type_by_name[card_type_name]
        return QtWidgets.QDialog.accept(self)



#
# renderer.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.component import Component


class Renderer(Component):

    """Assembles a sequence of 'fact_keys' which are keys in a dictionary
    'fact_data' to a certain format, e.g. a html page, or a purely text based
    format.
    
    Typically 'fact_keys' are the question and answer keys of the card's fact
    view.

    It is contained in a RenderChain and represents the functionality which
    is typically different for each card type.

    If this renderer is only for a specific card type (and its descendants)
    'used_for' can be set to the corresponding CardType class. If it is set to
    None, this renderer is used for all other card types.

    """

    component_type = "renderer"
    used_for = None  # Used for all card types.

    def update(self, card_type):

        """Update renderer information for given card type. Some information
        (e.g. css style sheets) is typically cached, and this function is
        used to signal that the cache should be rebuilt.

        """
        
        pass

    def render(self, fact_data, fact_keys, card_type, **render_args):

        """Assembles a sequence of 'fact_keys' which are keys in a dictionary
        'fact_data'.
        
        card_type' is passed as extra argument e.g. to determine card type
        specific formatting.

        """
        
        raise NotImplementedError



#
# test_activate_cards.py <Peter.Bienstman@UGent.be>
#

from nose.tools import raises
from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne import Mnemosyne
from mnemosyne.libmnemosyne.criterion import Criterion
from mnemosyne.libmnemosyne.criteria.default_criterion import DefaultCriterion

class TestActivateCards(MnemosyneTest):

    def test_compare_criteria(self):
        c1 = DefaultCriterion(self.mnemosyne.component_manager)
        c1.deactivated_card_type_fact_view_ids = set()
        c1._tag_ids_active = set([self.database().get_or_create_tag_with_name("default")._id])
        c1._tag_ids_forbidden = set()

        c2 = DefaultCriterion(self.mnemosyne.component_manager)
        c2.deactivated_card_type_fact_view_ids = set()
        c2._tag_ids_active = set([self.database().get_or_create_tag_with_name("default")._id])
        c2._tag_ids_forbidden = set()

        c3 = DefaultCriterion(self.mnemosyne.component_manager)
        c3.deactivated_card_type_fact_view_ids = set()
        c3._tag_ids_active = set([self.database().get_or_create_tag_with_name("default1")._id])
        c3._tag_ids_forbidden = set()

        assert c1 == c2
        assert c1 != c3
        assert c1 != 1

    def test_activate_cards_1(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
                                              grade=-1, tag_names=["default"])
        assert self.database().active_count() == 1

        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_2 = self.card_type_with_id("2")
        self.controller().create_new_cards(fact_data, card_type_2,
                                              grade=-1, tag_names=["default"])
        assert self.database().active_count() == 3

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("default")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 3

        c.deactivated_card_type_fact_view_ids = \
            set([(card_type_2.id, card_type_2.fact_views[0].id)])
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("default")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 2

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("default2")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

        fact_data = {"f": "question2",
                     "b": "answer2"}
        self.controller().create_new_cards(fact_data, card_type_2,
                                              grade=-1, tag_names=["default2"])
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("default2")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 2

        fact_data = {"f": "question3",
                     "b": "answer3"}
        self.controller().create_new_cards(fact_data, card_type_2,
                                              grade=-1, tag_names=["default3",
                                                                  "default4"])
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("default3")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 2

        c.deactivated_card_type_fact_view_ids = \
            set([(card_type_2.id, card_type_2.fact_views[0].id)])
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("default3")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 1

        c.deactivated_card_type_fact_view_ids = \
            set([(card_type_2.id, card_type_2.fact_views[0].id),
                 (card_type_2.id, card_type_2.fact_views[1].id)])
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("default3")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

    def test_activate_cards_2(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["active", "forbidden"])
        assert self.database().active_count() == 1

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("active")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 1

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("active")._id])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("forbidden")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

    def test_activate_cards_3(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["tag1"])[0]
        self.review_controller().show_new_question()

        assert self.review_controller().card == card
        assert self.review_controller().counters() == (0, 1, 1)

        fact_data = {"f": "question2",
                     "b": "answer2"}
        card_type_2 = self.card_type_with_id("2")
        cards = self.controller().create_new_cards(fact_data, card_type_2,
           grade=-1, tag_names=["tag2"])

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("tag2")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        self.review_controller().reset_but_try_to_keep_current_card()
        self.review_controller().reload_counters()

        assert self.review_controller().card != card
        assert self.review_controller().counters() == (0, 2, 2)

    def test_activate_cards_4(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["tag"])[0]
        self.review_controller().show_new_question()

        assert self.review_controller().card == card
        assert self.review_controller().counters() == (0, 1, 1)

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("tag2")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        self.review_controller().reset_but_try_to_keep_current_card()
        self.review_controller().reload_counters()

        assert self.review_controller().card is None
        assert self.review_controller().counters() == (0, 0, 0)

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("tag")._id])
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        self.review_controller().reset_but_try_to_keep_current_card()
        self.review_controller().reload_counters()

        assert self.review_controller().card == card
        assert self.review_controller().counters() == (0, 1, 1)

    def test_activate_cards_new(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["forbidden"])
        assert self.database().active_count() == 1

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("forbidden")._id])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("forbidden")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

        fact_data = {"f": "question2",
                     "b": "answer2"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["forbidden"])
        assert self.database().active_count() == 0

    def test_activate_cards_new_2(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["default"])
        assert self.database().active_count() == 1

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = \
            set([(card_type_1.id, card_type_1.fact_views[0].id)])
        c._tag_ids_active = set()
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

        fact_data = {"f": "question2",
                     "b": "answer2"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["default"])
        assert self.database().active_count() == 0

    def test_activate_cards_edit(self):
        fact_data = {"f": "question3",
                     "b": "answer3"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["forbidden"])[0]
        assert self.database().active_count() == 1

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("forbidden")._id])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("forbidden")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

        card_type_2 = self.card_type_with_id("2")
        self.controller().edit_card_and_sisters(card, card.fact.data,
               card_type_2, new_tag_names=["allowed"], correspondence=[])
        assert self.database().active_count() == 2

        c = list(self.database().criteria())[0]
        assert len(c._tag_ids_forbidden) == 0
        assert len(c._tag_ids_active) == 1

    def test_activate_cards_bulk_edit(self):
        fact_data = {"f": "question3",
                     "b": "answer3"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["allowed"])[0]
        assert self.database().active_count() == 1

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("allowed")._id])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("forbidden")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 1

        self.database().add_tag_to_cards_with_internal_ids(\
            self.database().get_or_create_tag_with_name("forbidden"), [card._id])
        
        assert self.database().active_count() == 0

    def test_activate_cards_bulk_edit_2(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["forbidden"])[0]

        fact_data = {"f": "question2",
                     "b": "answer2"}
        card = self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["forbidden"])[0]
        assert self.database().active_count() == 2

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("__UNTAGGED__")._id])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("forbidden")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

        self.database().remove_tag_from_cards_with_internal_ids(\
            self.database().get_or_create_tag_with_name("forbidden"), [card._id])

        assert self.database().active_count() == 1
        
    def test_card_type(self):
        from mnemosyne.libmnemosyne.card_types.cloze import ClozePlugin
        for plugin in self.plugins():
            if isinstance(plugin, ClozePlugin):
                cloze_plugin = plugin
                plugin.activate()
                break

        fact_data = {"text": "[foo]"}
        card_type_1 = self.card_type_with_id("5")
        card = self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["default"])[0]
        assert self.database().active_count() == 1

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = \
            set([(card_type_1.id, card_type_1.fact_views[0].id)])
        c._tag_ids_active = set()
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

        self.controller().delete_facts_and_their_cards([card.fact])
        plugin.deactivate()
        c = self.database().current_criterion()
        assert len(c.deactivated_card_type_fact_view_ids) == 0

    def test_cached_scheduler_count(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["tag"])
        assert self.database().active_count() == 1
        self.review_controller().show_new_question()
        assert self.review_controller().active_count == 1

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set()
        c._tag_ids_forbidden = set()
        self.database().set_current_criterion(c)
        self.review_controller().reset_but_try_to_keep_current_card()
        assert self.database().active_count() == 0
        assert self.review_controller().active_count == 0

    def test_activate_cards_5(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["b"])

        fact_data = {"f": "question2",
                     "b": "answer2"}
        self.controller().create_new_cards(fact_data, card_type_1,
            grade=-1, tag_names=["a"])
        assert self.database().active_count() == 2

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("a")._id,
                                 self.database().get_or_create_tag_with_name("b")._id])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("b")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 1

    @raises(AssertionError)
    def test_activate_cards_6(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["a"])

        fact_data = {"f": "question2",
                     "b": "answer2"}
        self.controller().create_new_cards(fact_data, card_type_1,
            grade=-1, tag_names=["b"])
        assert self.database().active_count() == 2

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("b")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 1

    def test_empty_criterion(self):
        c = Criterion(self.mnemosyne.component_manager)
        assert c.is_empty() == False

        tag_id = self.database().get_or_create_tag_with_name("a")._id

        c = DefaultCriterion(self.mnemosyne.component_manager)
        for card_type in self.card_types():
            for fact_view in card_type.fact_views:
                c.deactivated_card_type_fact_view_ids.add((card_type.id, fact_view.id))
        c._tag_ids_active = set([])
        c._tag_ids_forbidden = set()
        assert c.is_empty() == True

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([])
        c._tag_ids_forbidden = set()
        assert c.is_empty() == True

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set(["__UNTAGGED__"])
        c._tag_ids_forbidden = set([tag_id])
        assert c.is_empty() == False

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set(["__UNTAGGED__"])
        c._tag_ids_forbidden = set([tag_id, "__UNTAGGED__"])
        assert c.is_empty() == True

    def test_inactive_parent(self):

        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["dummy"])

        fact_data = {"f": "question2",
                     "b": "answer2"}
        card_type_1 = self.card_type_with_id("1")
        self.controller().create_new_cards(fact_data, card_type_1,
           grade=-1, tag_names=["a"])

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("dummy")._id])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("a")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 1

        fact_data = {"f": "question3",
                     "b": "answer3"}
        self.controller().create_new_cards(fact_data, card_type_1,
            grade=-1, tag_names=["a::b"])

        assert self.database().active_count() == 1
        fact_data = {"f": "question4",
                     "b": "answer4"}
        self.controller().create_new_cards(fact_data, card_type_1,
            grade=-1, tag_names=["dummy::b"])
        assert self.database().active_count() == 2



#
# plugin.py <Peter.Bienstman@UGent.be>
#

import importlib

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.component import Component


class Plugin(Component):

    """A Plugin is a group of components which can be activated and
    deactivated by the user when the program is running.

    Plugins are used as a separate class, and not as mixins, in order to
    allow for the fact that a Plugin can group multiple components.

    'components' is a list of component classes (not instances) that will
    be registered and/or instantiated when the Plugin becomes active.

    'gui_for_component' is a dictionary with as key a component class to which
    a GUI component needs to be added. The value is a list of tuples
    [("module_name", "class_name")]

    Activating and deactivating certain components needs to give rise to
    certain side effects. It's cumbersone to implement those in the
    'activate' and 'deactivate' methods of the components, as these also
    are called when the program is still starting up and the context can
    be completely different. Therefore, they are handled here.

    """

    name = ""
    description = ""
    component_type = "plugin"
    components = []
    gui_for_component = {}
    current_API_level = 3

    def __init__(self, component_manager):
        Component.__init__(self, component_manager)
        assert self.name and self.description, \
            "A plugin needs a name and a description."
        self.instantiated_components = []
        self.registered_components = []
        self.review_reset_needed = False

    def activate(self):
        if not hasattr(self, "supported_API_level") or \
            self.supported_API_level < self.current_API_level:
            self.main_widget().show_error(_("Plugin '" + self.name +\
                "' needs to be upgraded. Please contact its author."))
            return
        # Don't activate a plugin twice.
        if self.instantiated_components or self.registered_components:
            return
        # See if we need to reset the review process.
        self.review_reset_needed = False
        for component in self.components:
            if component.component_type in \
                ["scheduler", "review_controller", "review_widget"]:
                self.review_reset_needed = True
        # Register all our regular components. Instantiate them if needed.
        for component in self.components:
            if component.instantiate == Component.IMMEDIATELY:
                component = component(component_manager=self.component_manager)
                self.component_manager.register(component)
                component.activate()
                self.instantiated_components.append(component)
            else:
                self.component_manager.register(component)
                self.registered_components.append(component)
        # Register gui components for certain classes.
        for key, value in self.gui_for_component.items():
            component_name = key
            assert type(component_name) == str
            for gui_module_name, gui_class_name in value:
                gui_class = getattr(\
                    importlib.import_module(gui_module_name), gui_class_name)
                self.component_manager.add_gui_to_component(\
                    component_name, gui_class)
        # Make necessary side effects happen.
        for component in self.components:
            if component.used_for == "configuration_defaults":
                component(self.component_manager).run()
        db = self.database()
        if db and db.is_loaded():
            for component in self.instantiated_components:
                if component.component_type == "card_type":
                    for criterion in db.criteria():
                        criterion.active_card_type_added(component)
                        db.update_criterion(criterion)
        if self.database().is_loaded() and self.review_reset_needed:
            self.controller().reset_study_mode()
            # We need to log 'started_scheduler' events manually and not
            # from e.g. the 'activate' function of the scheduler because
            # during startup, the database is not yet loaded when the
            # scheduler gets activated.
            self.log().started_scheduler()
        # Use names instead of instances here in order to survive pickling.
        self.config()["active_plugins"].add(self.__class__.__name__)

    def deactivate(self):
        db = self.database()
        # Check if we are allowed to deactivate a card type.
        can_deactivate = True
        if db and db.is_loaded():
            for component in self.instantiated_components:
                if component.component_type == "card_type":
                    if self.database().has_clones(component):
                        can_deactivate = False
                    for card_type in self.database().card_types_in_use():
                        if issubclass(card_type.__class__,
                                      component.__class__):
                            can_deactivate = False
                            break
                    if can_deactivate == False:
                        self.main_widget().show_error(_("Cannot deactivate") \
                            + " '" + component.name + "'. " + \
_("There are cards with this card type (or a clone of it) in the database."))
                        return False
                    for criterion in db.criteria():
                        criterion.card_type_deleted(component)
                        db.update_criterion(criterion)
        # Deactivate and unregister components.
        for component in self.instantiated_components:
            component.deactivate()
            self.component_manager.unregister(component)
        for component in self.registered_components:
            self.component_manager.unregister(component)
        # If we are shutting down and the database is gone, don't worry about
        # side effects.
        if not self.database():
            return True
        # Make necessary side effects happen.
        if self.review_reset_needed:
            self.controller().reset_study_mode()
            self.log().started_scheduler()
        # Use names instead of instances here in order to survive pickling.
        if self.__class__.__name__ in self.config()["active_plugins"]:
            self.config()["active_plugins"].remove(self.__class__.__name__)
        self.instantiated_components = []
        self.registered_components = []
        return True


def register_user_plugin(plugin_class):

    """Plugins defined in the user's plugin directory don't have an easy
    access to the current component manager, which is why this convenience
    function is provided. User defined plugins only make sense in a single
    user context (a.o. because of security reasons), so we can just take the
    first registered component manager as being the correct one.

    """

    from .component_manager import _component_managers
    key = list(_component_managers.keys())[0]
    component_manager = _component_managers[key]
    plugin = plugin_class(component_manager=component_manager)
    component_manager.register(plugin)
    return plugin



#
# mnemosyne1_xml.py <Johannes.Baiter@gmail.com>
#

import os
import time
from xml.etree import cElementTree

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.file_format import FileFormat
from mnemosyne.libmnemosyne.utils import rand_uuid, MnemosyneError
from mnemosyne.libmnemosyne.file_formats.mnemosyne1 import Mnemosyne1


class Mnemosyne1XML(FileFormat, Mnemosyne1):

    description = _("Mnemosyne 1.x *.XML files")
    extension = ".xml"
    filename_filter = _("Mnemosyne 1.x XML files") + " (*.xml)"
    import_possible = True
    export_possible = False

    def __init__(self, component_manager):
        FileFormat.__init__(self, component_manager)
        Mnemosyne1.__init__(self, component_manager)
        self.anon_to_id = {}

    def do_import(self, filename, extra_tag_names=""):
        FileFormat.do_import(self, filename, extra_tag_names)
        w = self.main_widget()
        # The import process generates card log entries which have new 2.0
        # ids as opposed to their old 1.x ids, so we need to delete them
        # later.
        db = self.database()
        log_index = db.current_log_index()
        try:
            w.set_progress_text(_("Importing cards..."))
            self.read_items_from_mnemosyne1_xml(filename)
            self.create_cards_from_mnemosyne1(extra_tag_names)
        except MnemosyneError:
            w.close_progress()
            return
        db.remove_card_log_entries_since(log_index)
        # We now generate 'added card' events with the proper ids.
        timestamp = int(time.time())
        for item in self.items:
            db.log_added_card(timestamp, item.id)
        self.database().link_inverse_cards()
        w.close_progress()
        self.warned_about_missing_media = False

    def read_items_from_mnemosyne1_xml(self, filename):
        # Reset anonymiser when importing a new file, otherwise information
        # from the previous file still lingers and we get erroneously think
        # we've imported this before.
        self.anon_to_id = {}
        w = self.main_widget()
        try:
            tree = cElementTree.parse(filename)
        except cElementTree.ParseError as e:
            w.show_error(_("Unable to parse file:") + str(e))
            raise MnemosyneError
        except:
            w.show_error(_("Unable to open file."))
            raise MnemosyneError
        if tree.getroot().tag != "mnemosyne" or \
            tree.getroot().get("core_version") != "1":
            w.show_error(_
                    ("XML file does not seem to be a Mnemosyne 1.x XML file."))
            raise MnemosyneError
        self.starttime = 0
        if tree.getroot().get("time_of_start"):
            self.starttime = int(tree.getroot().get("time_of_start"))
        category_with_name = {}
        self.categories = []
        for element in tree.findall("category"):
            category = Mnemosyne1.MnemosyneCore.Category()
            category.name = element.find("name").text
            category.active = bool(element.get("active"))
            self.categories.append(category)
            category_with_name[category.name] = category
        self.items = []
        warned_about_import = False
        for element in tree.findall("item"):
            item = Mnemosyne1.MnemosyneCore.Item()
            item.id = element.get("id")
            if not item.id:
                item.id = rand_uuid()
            if item.id.startswith('_'):
                item.id = self.unanonymise_id(item.id)
            item.q = element.find("Q").text
            item.a = element.find("A").text
            if item.a is None:
                item.a = ""
            item.cat = category_with_name[element.find("cat").text]
            if element.get("gr"):
                if not warned_about_import:
                    result = w.show_question(_("This XML file contains learning data. It's best to import this from a mem file, in order to preserve historical statistics. Continue?"), _("Yes"), _("No"), "")
                    warned_about_import = True
                    if result == 1:  # No
                        return
                item.grade = int(element.get("gr"))
            else:
                item.grade = 0
            if element.get("e"):
                item.easiness = float(element.get("e"))
            else:
                item.easiness = 2.5
            if element.get("ac_rp"):
                item.acq_reps = int(element.get("ac_rp"))
            else:
                item.acq_reps = 0
            if element.get("rt_rp"):
                item.ret_reps = int(element.get("rt_rp"))
            else:
                item.ret_reps = 0
            if element.get("lps"):
                item.lapses = int(element.get("lps"))
            else:
                item.lapses = 0
            if element.get("ac_rp_l"):
                item.acq_reps_since_lapse = int(element.get("ac_rp_l"))
            else:
                item.acq_reps_since_lapse = 0
            if element.get("rt_rp_l"):
                item.ret_reps_since_lapse = int(element.get("rt_rp_l"))
            else:
                item.ret_reps_since_lapse = 0
            if element.get("l_rp"):
                item.last_rep = int(float(element.get("l_rp")))
            else:
                item.last_rep = 0
            if element.get("n_rp"):
                item.next_rep = int(float(element.get("n_rp")))
            else:
                item.next_rep = 0
            if element.get("u"):
                item.unseen = bool(element.get("u"))
            else:
                if item.acq_reps <= 1 and item.ret_reps == 0 \
                    and item.grade == 0:
                    item.unseen = True
                else:
                    item.unseen = False
            self.items.append(item)

    def unanonymise_id(self, item_id):
        if "." in item_id:
            old_id, suffix = item_id.split(".", 1)
            suffix = "." + suffix
        else:
            old_id, suffix = item_id, ""
        if old_id in self.anon_to_id:
            item_id = self.anon_to_id[old_id]
        else:
            item_id = rand_uuid()
            self.anon_to_id[old_id] = item_id
        return item_id + suffix




#
# french.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.language import Language


class French(Language):

    name = _("French")
    used_for = "fr"
    sublanguages = {"fr-FR": _("French (France)"),
                     "fr-CA": _("French (Canada)")}
    feature_description = _("Google translation and text-to-speech.")



#
# configuration_wdgt_study.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.ui_components.configuration_widget import \
     ConfigurationWidget
from mnemosyne.pyqt_ui.ui_configuration_wdgt_study import \
     Ui_ConfigurationWdgtStudy
from mnemosyne.libmnemosyne.schedulers.cramming import RANDOM, \
    EARLIEST_FIRST, LATEST_FIRST, MOST_LAPSES_FIRST


class ConfigurationWdgtStudy(QtWidgets.QWidget, ConfigurationWidget,
    Ui_ConfigurationWdgtStudy):

    name = _("Study")

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.setupUi(self)
        if self.config()["randomise_scheduled_cards"] == True:
            self.scheduled_cards.setCurrentIndex(1)
        else:
            self.scheduled_cards.setCurrentIndex(0)
        self.non_memorised_cards.setValue(self.config()\
            ["non_memorised_cards_in_hand"])
        if self.config()["randomise_new_cards"] == True:
            self.new_cards.setCurrentIndex(1)
        else:
            self.new_cards.setCurrentIndex(0)
        if self.config()["cramming_order"] == RANDOM:
            self.order.setCurrentIndex(0)
        elif self.config()["cramming_order"] == EARLIEST_FIRST:
            self.order.setCurrentIndex(1)
        elif self.config()["cramming_order"] == LATEST_FIRST:
            self.order.setCurrentIndex(2)
        elif self.config()["cramming_order"] == MOST_LAPSES_FIRST:
            self.order.setCurrentIndex(3)
        self.max_ret_reps_for_recent_cards.setValue(self.config()\
            ["max_ret_reps_for_recent_cards"])
        self.max_ret_reps_since_lapse.setValue(self.config()\
            ["max_ret_reps_since_lapse"])
        if self.config()["cramming_store_state"] == True:
            self.store_state.setCheckState(QtCore.Qt.Checked)
        else:
            self.store_state.setCheckState(QtCore.Qt.Unchecked)

    def reset_to_defaults(self):
        answer = self.main_widget().show_question(\
            _("Reset current tab to defaults?"), _("&Yes"), _("&No"), "")
        if answer == 1:
            return
        self.scheduled_cards.setCurrentIndex(0)
        self.non_memorised_cards.setValue(10)
        self.new_cards.setCurrentIndex(0)
        self.order.setCurrentIndex(0)
        self.max_ret_reps_for_recent_cards.setValue(1)
        self.max_ret_reps_since_lapse.setValue(999999)
        self.store_state.setCheckState(QtCore.Qt.Checked)

    def apply(self):
        if self.scheduled_cards.currentIndex() == 1:
            self.config()["randomise_scheduled_cards"] = True
        else:
            self.config()["randomise_scheduled_cards"] = False
        self.config()["non_memorised_cards_in_hand"] = \
            self.non_memorised_cards.value()
        if self.new_cards.currentIndex() == 1:
            self.config()["randomise_new_cards"] = True
        else:
            self.config()["randomise_new_cards"] = False
        if self.order.currentIndex() == 0:
            self.config()["cramming_order"] = RANDOM
        elif self.order.currentIndex() == 1:
            self.config()["cramming_order"] = EARLIEST_FIRST
        elif self.order.currentIndex() == 2:
            self.config()["cramming_order"] = LATEST_FIRST
        elif self.order.currentIndex() == 3:
            self.config()["cramming_order"] = MOST_LAPSES_FIRST
        self.config()["max_ret_reps_for_recent_cards"] = \
            self.max_ret_reps_for_recent_cards.value()
        self.config()["max_ret_reps_since_lapse"] = \
            self.max_ret_reps_since_lapse.value()
        if self.store_state.checkState() == QtCore.Qt.Checked:
            self.config()["cramming_store_state"] = True
        else:
            self.config()["cramming_store_state"] = False





#
# translator.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.component import Component


class Translator(Component):

    """Generic translation service for words and sentences. Not to be confused
    the GuiTranslator, which is for translating the user interface.

    Note that a single component can handle multiple languages (e.g. Google TTS)
    and the language to be used will be determined through the language_id
    property of the card_type argument.

    """

    component_type = "translator"
    used_for = None # Single ISO 639-1 code, or multiple as tuple of strings.
    popup_menu_text = None # "Insert translation..."

    def translate(self, card_type, foreign_text, dest_language_id):

        """Returns translated text."""

        raise NotImplementedError

    def show_dialog(self, card_type, foreign_text):

        """Returns translated text to insert. The user can set the target
        language in the dialog."""

        dialog = self.gui_components[0](\
            translator=self, component_manager=self.component_manager)
        self.component_manager.register(dialog)
        dialog.activate(card_type, foreign_text)
        self.instantiated_gui_components.append(dialog)
        return dialog.text_to_insert



#
# manage_plugins_dlg.py <Peter.Bienstman@UGent.be>
#

import os
from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.pyqt_ui.ui_manage_plugins_dlg import Ui_ManagePluginsDlg
from mnemosyne.libmnemosyne.ui_components.dialogs import ManagePluginsDialog


class ManagePluginsDlg(QtWidgets.QDialog, ManagePluginsDialog,
                       Ui_ManagePluginsDlg):

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.setupUi(self)
        self.setWindowFlags(self.windowFlags() \
            | QtCore.Qt.WindowMinMaxButtonsHint)
        self.setWindowFlags(self.windowFlags() \
            & ~ QtCore.Qt.WindowContextHelpButtonHint)
        self.last_selected_row = 0
        self.build_plugin_list()
        state = self.config()["plugins_dlg_state"]
        if state:
            self.restoreGeometry(state)

    def build_plugin_list(self):
        plugin_dir = os.path.join(self.config().data_dir, "plugins")
        self.can_be_deleted = [filename.rsplit(".", 1)[0] for \
            filename in os.listdir(plugin_dir) \
            if filename.endswith(".manifest")]
        self.plugin_list.clear()
        self.previously_active = {}
        self.plugin_with_name = {}
        for plugin in self.plugins():
            list_item = QtWidgets.QListWidgetItem(_(plugin.name))
            list_item.setFlags(list_item.flags() \
                | QtCore.Qt.ItemIsUserCheckable)
            self.plugin_with_name[_(plugin.name)] = plugin
            active = \
                plugin.__class__.__name__ in self.config()["active_plugins"]
            self.previously_active[_(plugin.name)] = active
            if active:
                list_item.setCheckState(QtCore.Qt.Checked)
            else:
                list_item.setCheckState(QtCore.Qt.Unchecked)
            self.plugin_list.addItem(list_item)
        self.plugin_list.setCurrentRow(self.last_selected_row)
        self.plugin_description.setText(_(self.plugins()[0].description))
        self.delete_button.setEnabled(\
            self.plugins()[0].__class__.__name__ in self.can_be_deleted)

    def plugin_selected(self, list_item, dummy=None):
        if list_item is None: # Sometimes Qt fires spurious events.
            return
        # If we get there through activating of the item, make sure we move
        # the selection to the activated item.
        self.plugin_list.setCurrentItem(list_item)
        plugin = self.plugin_with_name[list_item.text()]
        self.plugin_description.setText(_(plugin.description))
        self.delete_button.setEnabled(\
            plugin.__class__.__name__ in self.can_be_deleted)

    def activate(self):
        self.exec_()

    def _store_state(self):
        self.config()["plugins_dlg_state"] = self.saveGeometry()

    def closeEvent(self, event):
        # Generated when clicking the window's close button.
        self._store_state()

    def accept(self):
        # 'accept' does not generate a close event.
        self._store_state()
        for index in range(self.plugin_list.count()):
            list_item = self.plugin_list.item(index)
            plugin_name = list_item.text()
            if list_item.checkState() == QtCore.Qt.Checked and \
                self.previously_active[plugin_name] == False:
                self.plugin_with_name[plugin_name].activate()
            elif list_item.checkState() == QtCore.Qt.Unchecked and \
                self.previously_active[plugin_name] == True:
                self.plugin_with_name[plugin_name].deactivate()
        return QtWidgets.QDialog.accept(self)

    def install_plugin(self):
        self.last_selected_row = self.plugin_list.currentRow()
        self.controller().install_plugin()
        self.build_plugin_list()

    def delete_plugin(self):
        plugin_name = self.plugin_list.selectedItems()[0].text()
        question = _("Are you sure you want to delete the plugin") + " \"" + \
            plugin_name + "\" " + _("and not just deactivate it?")
        answer = self.main_widget().show_question(question,
            _("Delete"), _("Cancel"), "")
        if answer == 1:  # Cancel
            return
        self.controller().delete_plugin(self.plugin_with_name[plugin_name])
        self.build_plugin_list()



#
# test_statistics.py <Peter.Bienstman@UGent.be>
#

import os
import time
import datetime

from nose.tools import raises

HOUR = 60 * 60 # Seconds in an hour.
DAY = 24 * HOUR # Seconds in a day.

from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne.file_formats.science_log_parser import ScienceLogParser

from openSM2sync.log_entry import EventTypes

class TestStatistics(MnemosyneTest):

    def test_current_card(self):
        from mnemosyne.libmnemosyne.statistics_pages.current_card import CurrentCard
        page = CurrentCard(self.mnemosyne.component_manager)
        page.prepare_statistics(0)
        assert "No current card." in page.html
        card_type = self.card_type_with_id("2")
        fact_data = {"f": "f", "b": "b"}
        card_1, card_2 = self.controller().create_new_cards(fact_data,
          card_type, grade=-1, tag_names=["default"])
        self.review_controller().show_new_question()
        assert self.database().card_count_for_fact_view\
               (card_type.fact_views[0], active_only=True) == 1
        page.prepare_statistics(0)
        assert "Unseen card, no statistics available yet." in page.html
        self.review_controller().grade_answer(1)
        page.prepare_statistics(0)
        assert "No current card." not in page.html
        assert "Unseen card, no statistics available yet." not in page.html

    def test_easiness(self):
        from mnemosyne.libmnemosyne.statistics_pages.easiness import Easiness
        page = Easiness(component_manager=self.mnemosyne.component_manager)
        page.prepare_statistics(-1)
        assert len(page.data) == 0

        card_type = self.card_type_with_id("2")
        fact_data = {"f": "f", "b": "b"}
        card_1, card_2 = self.controller().create_new_cards(fact_data,
          card_type, grade=-1, tag_names=["default"])
        self.review_controller().show_new_question()
        self.review_controller().grade_answer(1)
        page = Easiness(component_manager=self.mnemosyne.component_manager)

        page.prepare_statistics(-1)
        assert page.data == [2.5]
        page.prepare_statistics(page.variants[1][0])
        assert page.data == [2.5]

    @MnemosyneTest.set_timezone_utc
    def test_past_schedule(self):
        self.database().update_card_after_log_import = (lambda x, y, z: 0)
        self.database().before_1x_log_import()
        filename = os.path.join(os.getcwd(), "tests", "files", "schedule_1.txt")
        ScienceLogParser(self.database()).parse(filename)
        days_elapsed = datetime.date.today() - datetime.date(2009, 8, 15)
        assert self.scheduler().card_count_scheduled_n_days_from_now(\
            -days_elapsed.days) == 124
        assert self.scheduler().card_count_scheduled_n_days_from_now(-1) == 0

    def test_past_schedule_2_machines(self):
        con = self.database().con
        con.execute("""insert into log(event_type, timestamp, object_id,
            acq_reps,ret_reps, lapses) values(?,?,?,?,?,?)""",
            (EventTypes.LOADED_DATABASE, time.time() - DAY,
            "A", 20, -666, -666))
        con.execute("""insert into log(event_type, timestamp, object_id,
            acq_reps,ret_reps, lapses) values(?,?,?,?,?,?)""",
            (EventTypes.LOADED_DATABASE, time.time() - DAY,
            "B", 40, -666, -666))
        assert self.scheduler().card_count_scheduled_n_days_from_now(-10) == 0
        assert self.scheduler().card_count_scheduled_n_days_from_now(-1) == 20

    def test_past_schedule_extrapolated(self):
        con = self.database().con
        con.execute("""insert into log(event_type, timestamp, object_id,
            acq_reps,ret_reps, lapses) values(?,?,?,?,?,?)""",
            (EventTypes.LOADED_DATABASE, time.time() - DAY,
            "A.fut", 20, -666, -666))
        assert self.scheduler().card_count_scheduled_n_days_from_now(-10) == 0
        assert self.scheduler().card_count_scheduled_n_days_from_now(-1) == 20

    def test_schedule_page(self):
        from mnemosyne.libmnemosyne.statistics_pages.schedule import Schedule
        page = Schedule(self.mnemosyne.component_manager)
        for i in range(1, 11):
            page.prepare_statistics(i)

    @raises(AttributeError)
    def test_schedule_page_2(self):
        from mnemosyne.libmnemosyne.statistics_pages.schedule import Schedule
        page = Schedule(self.mnemosyne.component_manager)
        page.prepare_statistics(0)

    @MnemosyneTest.set_timezone_utc
    def test_added_cards(self):
        self.database().update_card_after_log_import = (lambda x, y, z: 0)
        self.database().before_1x_log_import()
        filename = os.path.join(os.getcwd(), "tests", "files", "added_1.txt")
        ScienceLogParser(self.database()).parse(filename)
        days_elapsed = datetime.date.today() - datetime.date(2009, 8, 19)
        assert self.database().card_count_added_n_days_ago(days_elapsed.days) \
               == 2
        assert self.scheduler().card_count_scheduled_n_days_from_now(1) == 0

    def test_added_cards_page(self):
        from mnemosyne.libmnemosyne.statistics_pages.cards_added import CardsAdded
        page = CardsAdded(self.mnemosyne.component_manager)
        for i in range(1, 6):
            page.prepare_statistics(i)

    def test_learned_cards_page(self):
        from mnemosyne.libmnemosyne.statistics_pages.cards_learned import CardsLearned
        page = CardsLearned(self.mnemosyne.component_manager)
        for i in range(1, 6):
            page.prepare_statistics(i)

    @raises(AttributeError)
    def test_added_cards_page_2(self):
        from mnemosyne.libmnemosyne.statistics_pages.cards_added import CardsAdded
        page = CardsAdded(self.mnemosyne.component_manager)
        page.prepare_statistics(0)

    @MnemosyneTest.set_timezone_utc
    def test_score(self):
        self.database().update_card_after_log_import = (lambda x, y, z: 0)
        self.database().before_1x_log_import()
        filename = os.path.join(os.getcwd(), "tests", "files", "score_1.txt")
        ScienceLogParser(self.database()).parse(filename)
        days_elapsed = datetime.date.today() - datetime.date(2009, 8, 17)
        assert self.database().retention_score_n_days_ago(days_elapsed.days) \
               == 5/7.*100
        assert self.database().retention_score_n_days_ago(0) == 0
        from mnemosyne.libmnemosyne.statistics_pages.retention_score import RetentionScore
        page = RetentionScore(self.mnemosyne.component_manager)
        for i in range(1, 6):
            page.prepare_statistics(i)

    @raises(AttributeError)
    def test_score_page(self):
        from mnemosyne.libmnemosyne.statistics_pages.retention_score import RetentionScore
        page = RetentionScore(self.mnemosyne.component_manager)
        page.prepare_statistics(0)

    def test_card_count_for_tags(self):
        assert self.database().card_count_for_tags([], active_only=False) == 0


    def test_component_manager(self):

        from mnemosyne.libmnemosyne.statistics_page import HtmlStatisticsPage
        from mnemosyne.libmnemosyne.ui_components.statistics_widget import \
             StatisticsWidget

        class MyHtmlStatisticsWdgt(StatisticsWidget):
            used_for = HtmlStatisticsPage

        self.mnemosyne.component_manager.register(MyHtmlStatisticsWdgt)

        class MyPage(HtmlStatisticsPage):
            pass

        widget_class = self.mnemosyne.component_manager.current(\
                "statistics_widget", used_for=MyPage)

        assert widget_class == MyHtmlStatisticsWdgt



#
# backup.py <Peter.Bienstman@UGent.be>
#

import os
import sqlite3

from mnemosyne.libmnemosyne.hook import Hook
from mnemosyne.libmnemosyne.plugin import Plugin


class BackupHook(Hook):

    used_for = "after_backup"

    def run(self, backup_name):
        # Upload regular backup to a server.
        os.system("scp %s my.safe.server.com:" % backup_name)
        # Dump database to a text file (requires Python 2.6).
        with open('dump.sql', 'w') as f:
            for line in self.database().con.iterdump():
                f.write('%s\n' % line)


class BackupPlugin(Plugin):
    
    name = "Extra backup"
    description = "Move your backups to a safe place."   
    components = [BackupHook]
    supported_API_level = 3
      

# Register plugin.

from mnemosyne.libmnemosyne.plugin import register_user_plugin
register_user_plugin(BackupPlugin)





from .template import Template
import os.path
import re

class View:
    # Path where this view's template(s) live
    template_path = '.'

    # Extension for templates
    template_extension = 'mustache'

    # The name of this template. If none is given the View will try
    # to infer it based on the class name.
    template_name = None

    # Absolute path to the template itself. Pystache will try to guess
    # if it's not provided.
    template_file = None

    # Contents of the template.
    template = None

    # Character encoding of the template file. If None, Pystache will not
    # do any decoding of the template.
    template_encoding = None

    def __init__(self, template=None, context=None, **kwargs):
        self.template = template
        self.context = context or {}

        # If the context we're handed is a View, we want to inherit
        # its settings.
        if isinstance(context, View):
            self.inherit_settings(context)

        if kwargs:
            self.context.update(kwargs)

    def inherit_settings(self, view):
        """Given another View, copies its settings."""
        if view.template_path:
            self.template_path = view.template_path

        if view.template_name:
            self.template_name = view.template_name

    def load_template(self):
        if self.template:
            return self.template

        if self.template_file:
            return self._load_template()

        name = self.get_template_name() + '.' + self.template_extension

        if isinstance(self.template_path, str):
            self.template_file = os.path.join(self.template_path, name)
            return self._load_template()

        for path in self.template_path:
            self.template_file = os.path.join(path, name)
            if os.path.exists(self.template_file):
                return self._load_template()

        raise IOError('"%s" not found in "%s"' % (name, ':'.join(self.template_path),))


    def _load_template(self):
        f = open(self.template_file, 'r')
        try:
            template = f.read()
            if self.template_encoding:
                template = str(template, self.template_encoding)
        finally:
            f.close()
        return template

    def get_template_name(self, name=None):
        """TemplatePartial => template_partial
        Takes a string but defaults to using the current class' name or
        the `template_name` attribute
        """
        if self.template_name:
            return self.template_name

        if not name:
            name = self.__class__.__name__

        def repl(match):
            return '_' + match.group(0).lower()

        return re.sub('[A-Z]', repl, name)[1:]

    def __contains__(self, needle):
        return needle in self.context or hasattr(self, needle)

    def __getitem__(self, attr):
        val = self.get(attr, None)
        if not val:
            raise KeyError("No such key.")
        return val

    def get(self, attr, default):
        attr = self.context.get(attr, getattr(self, attr, default))

        if hasattr(attr, '__call__'):
            return attr()
        else:
            return attr

    def render(self, encoding=None):
        template = self.load_template()
        return Template(template, self).render(encoding=encoding)

    def __str__(self):
        return self.render()



#
# study_mode.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.component import Component


class StudyMode(Component):

    """A study mode is a collection of a scheduler and a review controller.
    Different study modes can share e.g. the same scheduler, but instantiated
    with different parameters.

    Each study more should also have a review widget, indicated through the
    'gui_for_component' mechanism.

    """

    id = ""
    name = ""
    menu_weight = 0 # To determine sorting order in menu
    component_type = "study_mode"
    Scheduler = None  # Class
    ReviewController = None  # Class

    def activate_components(self):
        # Register dependent classes first.
        self.component_manager.register(\
            self.Scheduler(self.component_manager))
        self.scheduler().activate()
        self.log().started_scheduler()
        self.component_manager.register(\
            self.ReviewController(self.component_manager))
        self.review_controller().activate()
        Component.activate_gui_components(self)

    def activate(self):
        self.activate_components()
        self.review_controller().reset()

    def deactivate(self):
        if self.scheduler():
            self.scheduler().deactivate()
            self.component_manager.unregister(self.scheduler())
        if self.review_controller():
            self.review_controller().deactivate()
            self.component_manager.unregister(self.review_controller())
        Component.deactivate(self)



#
# intro_wizard_dlg.py <Johannes.Baiter@gmail.com>
#

from PyQt5 import QtGui, QtCore, QtWidgets

from mnemosyne.pyqt_ui.ui_getting_started_dlg import Ui_GettingStartedDlg
from mnemosyne.libmnemosyne.ui_components.dialogs import GettingStartedDialog

class GettingStartedDlg(QtWidgets.QWizard, GettingStartedDialog, 
                        Ui_GettingStartedDlg):

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.setupUi(self)
        self.setWindowFlags(self.windowFlags() \
            | QtCore.Qt.WindowMinMaxButtonsHint)
        self.setWindowFlags(self.windowFlags() \
            & ~ QtCore.Qt.WindowContextHelpButtonHint)
        # Note: the svg file does not seem to work under windows.
        #watermark = QtGui.QPixmap(":/mnemosyne/pixmaps/mnemosyne.svg")\
        #    .scaledToHeight(200, QtCore.Qt.SmoothTransformation)
        watermark = QtGui.QPixmap(":/mnemosyne/pixmaps/mnemosyne.png")
        self.setPixmap(QtWidgets.QWizard.WatermarkPixmap, watermark)

    def activate(self):
        GettingStartedDialog.activate(self)
        self.exec_()

    def accept(self):
        self.config()["upload_science_logs"] = self.upload_box.isChecked()
        QtWidgets.QWizard.accept(self)




#
# import_rosetta_stone.py <Peter.Bienstman>
#

# Script to import audio and pictures from the Rosetta Stone V2 into a Sentence
# card type.

# Adapt it your own need. It uses linux external tools, so it needs to be
# modified to run under Windows.

import os
import shutil

from mnemosyne.script import Mnemosyne

# 'data_dir = None' will use the default sysem location, edit as appropriate.
data_dir = None
mnemosyne = Mnemosyne(data_dir)

# Answer questions coming from libmnemosyne.

def show_question(question, option0, option1, option2):
    # Aswer 'no' when adding duplicate cards.
    if question.startswith("There is already"):
        return 2
    # Answer 'yes' for making tag active.
    if question.startswith("Make tag"):
        return 0
    else:
        raise NotImplementedError

mnemosyne.main_widget().show_question = show_question

# This script will add tags like TRS Arabic::Unit 1::Lesson 1
tag_prefix = "TRS Arabic"

# Card type.
card_type = mnemosyne.card_type_with_id("6::Arabic MSA sentences")

# Directory containing foreign language, with directories like ARA01_01
# and PCT01_01
foreign_directory = "/home/pbienst/tmp/trs_arabic"

# Directory containing native language, to generate translations.
native_directory = "/home/pbienst/tmp/trs_english"

# Subdirectory in the media directory to used.
media_subdir = "trs_ara"
full_media_subdir = os.path.join(mnemosyne.database().media_dir(), media_subdir)
if not os.path.exists(full_media_subdir):
    os.mkdir(full_media_subdir)

# Codec that was used to encode the foreign language.
foreign_codec = "iso-8859-6"
native_codec = "latin-1"

# Extract txt.
def get_txt(directory, codec):
    txt = {}
    for path in sorted(os.listdir(directory)):
        subdir = os.path.join(directory, path)
        if os.path.isdir(subdir) and not path.startswith("PCT"):
            # Determine unit and lesson number.
            unit, lesson = path[3:].split("_")
            unit = int(unit)
            if unit not in txt:
                txt[unit] = {}
            lesson = int(lesson)
            # Determine sentences.
            txt_file = file(os.path.join(subdir,
                [x for x in os.listdir(subdir) if x.endswith(".TXT")][0]))
            entries = str(txt_file.read(), codec, errors="ignore") \
                .replace(chr(336), "\'") \
                .replace(chr(213), "\'") \
                .replace(chr(210), "\"") \
                .replace(chr(211), "\"") \
                .split("@")[1:-1]
            assert len(entries) == 40
            txt[unit][lesson] = entries
    return txt

foreign_txt = get_txt(foreign_directory, foreign_codec)
native_txt = get_txt(native_directory, native_codec)

# Extract images.
def extract_images(directory):
    images = {}
    for path in sorted(os.listdir(directory)):
        subdir = os.path.join(directory, path)
        if os.path.isdir(subdir) and path.startswith("PCT"):
            # Detemine unit and lesson number.
            unit, lesson = path[3:].split("_")
            unit = int(unit)
            if unit not in images:
                images[unit] = {}
            lesson = int(lesson)
            img_dir = os.path.join(subdir,
                [x for x in os.listdir(subdir) if x.startswith("P")][0])
            img_list = []
            for img in sorted(os.listdir(img_dir)):
                full_path = os.path.join(img_dir, img)
                if img.endswith("JPG"):
                    shutil.copyfile(full_path, os.path.join(full_media_subdir, img))
                    img_list.append(media_subdir + "/" + img)
                if img.endswith("PCT"):
                    os.system("convert " + full_path + " " + \
                        os.path.join(full_media_subdir, img).replace("PCT", "JPG"))
                    img_list.append(\
                        media_subdir + "/" + img.replace("PCT", "JPG"))
            images[unit][lesson] = img_list
    return images

images = extract_images(foreign_directory)

# Extract sound.
def extract_sound(directory):
    sound = {}
    for path in sorted(os.listdir(directory)):
        subdir = os.path.join(directory, path)
        if os.path.isdir(subdir) and not path.startswith("PCT"):
            # Determine unit and lesson number.
            unit, lesson = path[3:].split("_")
            unit = int(unit)
            if unit not in sound:
                sound[unit] = {}
            lesson = int(lesson)
            snd_dir = os.path.join(subdir,
                [x for x in os.listdir(subdir) if x.endswith("S")][0])
            snd_list = []
            for snd in sorted(os.listdir(snd_dir)):
                full_path = os.path.join(snd_dir, snd)
                if snd.endswith("SWA"):
                    os.system("mplayer -vo null -vc dummy -af resample=44100 -ao pcm:waveheader " \
                        + full_path + " && lame audiodump.wav " + \
                        os.path.join(full_media_subdir, snd).replace("SWA", "MP3"))
                    # High bitrate version, not really needed.
                    #os.system("mplayer -vo null -vc dummy -af resample=44100 -ao pcm:waveheader " \
                    #    + full_path + " && lame -h --resample 44.1 -b 128 audiodump.wav " + \
                    #    os.path.join(full_media_subdir, snd).replace("SWA", "MP3"))
                    snd_list.append(\
                        media_subdir + "/" + snd.replace("SWA", "MP3"))
            sound[unit][lesson] = snd_list
    return sound

sound = extract_sound(foreign_directory)

for unit in foreign_txt:
    for lesson in foreign_txt[unit]:
        print(("unit", unit, "lesson", lesson))
        for i in range(40):
            print((foreign_txt[unit][lesson][i]))
            print((native_txt[unit][lesson][i].replace(chr(336), "\'")))
            print((images[unit][lesson][i]))
            print((sound[unit][lesson][i]))
            print()
            fact_data = {"f": "["+foreign_txt[unit][lesson][i] + "]",
                "p_1": "<audio src=\"" + sound[unit][lesson][i] + "\">",
                "m_1": native_txt[unit][lesson][i] + \
                    "\n<img src=\"" + images[unit][lesson][i] + "\">"}
            mnemosyne.controller().create_new_cards(fact_data,
            card_type, grade=-1, tag_names=[tag_prefix + "::Unit " + str(unit)\
                + "::Lesson " + str(lesson)])
        print()

mnemosyne.finalise()



#
# controller.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.component import Component


class Controller(Component):

    """A collection of logic used by the GUI.  The logic related to the
    review process is split out in a separated controller class, to
    allow that to be swapped out easily.

    There are two classes of functions here: 'show_XXX_dialog', which needs
    to be called by the GUI to set everything up to show a certain dialog,
    and then the other functions, which the implementation of the actual
    dialog can use to perform GUI-independent operations.

    See also 'How to write a new frontend' in the docs.

    """

    component_type = "controller"

    def heartbeat(self):

        """For code that needs to run periodically."""

        pass

    def update_title(self):
        raise NotImplementedError

    def show_add_cards_dialog(self):
        raise NotImplementedError

    def set_study_mode(self, study_mode):
        raise NotImplementedError

    def create_new_cards(self, fact_data, card_type, grade,
            tag_names, check_for_duplicates=True, save=True):
        raise NotImplementedError

    def show_edit_card_dialog(self):
        raise NotImplementedError

    def edit_card_and_sisters(self, card, new_fact_data, new_card_type,
            new_tag_names, correspondence):
        raise NotImplementedError

    def change_card_type(self, facts, old_card_type, new_card_type,
                         correspondence):

        """Note: all facts should have the same card type."""

        raise NotImplementedError

    def star_current_card(self):
        raise NotImplementedError

    def delete_current_card(self):
        raise NotImplementedError

    def delete_facts_and_their_cards(self, facts):
        raise NotImplementedError

    def reset_current_card(self):
        raise NotImplementedError

    def reset_facts_and_their_cards(self, facts):
        raise NotImplementedError

    def clone_card_type(self, card_type, clone_name):
        raise NotImplementedError

    def delete_card_type(self, card_type):
        raise NotImplementedError

    def show_new_file_dialog(self):
        raise NotImplementedError

    def show_open_file_dialog(self):
        raise NotImplementedError

    def save_file(self):
        raise NotImplementedError

    def show_save_file_as_dialog(self):
        raise NotImplementedError

    def show_insert_img_dialog(self, filter):

        """Filter contains the file dialog filter with the supported
        filetypes.

        """

        raise NotImplementedError

    def show_insert_sound_dialog(self, filter):

        """Filter contains the file dialog filter with the supported
        filetypes.

        """

        raise NotImplementedError

    def show_insert_video_dialog(self, filter):

        """Filter contains the file dialog filter with the supported
        filetypes.

        """

        raise NotImplementedError

    def show_browse_cards_dialog(self):
        raise NotImplementedError

    def show_manage_plugins_dialog(self):
        raise NotImplementedError

    def install_plugin(self):
        raise NotImplementedError

    def delete_plugin(self, plugin):
        raise NotImplementedError

    def show_manage_card_types_dialog(self):
        raise NotImplementedError

    def show_statistics_dialog(self):
        raise NotImplementedError

    def show_configuration_dialog(self):
        raise NotImplementedError

    def show_import_file_dialog(self):
        raise NotImplementedError

    def show_export_file_dialog(self):
        raise NotImplementedError

    def show_sync_dialog(self):
        raise NotImplementedError

    def show_tip_dialog(self):
        raise NotImplementedError

    def show_getting_started_dialog(self):
        raise NotImplementedError

    def sync(self, server, port, username, password, ui=None):
        raise NotImplementedError



#
# gui_translator.py <Peter.Bienstman@UGent.be>
#                   <Johannes.Baiter@gmail.com>
#

from mnemosyne.libmnemosyne.component import Component

_ = None

class GuiTranslator(Component):

    """Note: static text will be marked as translatable by
    `static_variable = _("...")`, e.g. in CardType. However, if want to have
    this text show up translated in the GUI, we need to use
    `_(static_variable)` there as opposed to just `static_variable`.

    This is the most straightforward design to allow switching back and forth
    between different languages in the GUI on the fly, i.e. without restarting
    the program.

    """

    component_type = "gui_translator"

    def __init__(self, component_manager):
        Component.__init__(self, component_manager)
        global _
        _ = self
        # We install a dummy translator so that translatable stings can be
        # marked even if 'activate' has not yet been called.
        self._translator = lambda x : x

    def activate(self):
        Component.activate(self)
        self.set_language(self.config()["ui_language"])
        # We cannot perform this in __init__, as we need to wait until
        # 'config' has been activated.

    def supported_languages(self):
        return []

    def set_language(self, language):

        """'language' should be an iso 693-1 code."""

        self.set_translator(language)
        self.translate_ui(language)

    def set_translator(self, language):
        raise NotImplementedError

    def translate_ui(self, language):

        """To be overridden by a GUI to do GUI-specific translation."""

        pass

    def __call__(self, text):

        """Used to do translations / mark translatable strings by _("...")."""

        return self._translator(text)


# http://en.wikipedia.org/wiki/List_of_ISO_639-2_codes

language_name_for_iso6931_code = {
    "aa": "Afar",
    "ab": "Abkhazian",
    "ae": "Avestan",
    "af": "Afrikaans",
    "ak": "Akan",
    "am": "Amharic",
    "an": "Aragonese",
    "ar": "Arabic",
    "as": "Assamese",
    "av": "Avaric",
    "ay": "Aymara",
    "az": "Azerbaijani",
    "ba": "Bashkir",
    "be": "Belarusian",
    "bg": "Bulgarian",
    "bh": "Bihari languages",
    "bi": "Bislama",
    "bm": "Bambara",
    "bn": "Bengali",
    "bo": "Tibetan",
    "br": "Breton",
    "bs": "Bosnian",
    "ca": "Catalan",
    "ca@valencia": "Catalan (Valencian)",
    "ce": "Chechen",
    "ch": "Chamorro",
    "co": "Corsican",
    "cr": "Cree",
    "cs": "Czech",
    "cu": "Church Slavonic, Church Slavic, Old Church Slavonic, Old Slavonic, Old Bulgarian",
    "cv": "Chuvash",
    "cy": "Welsh",
    "da": "Danish",
    "de": "German",
    "dv": "Divehi, Dhivehi, Maldivian",
    "dz": "Dzongkha",
    "ee": "Ewe",
    "el": "Modern Greek (1453-)",
    "en": "English",
    "eo": "Esperanto",
    "es": "Spanish (Castilian)",
    "et": "Estonian",
    "eu": "Basque",
    "fa": "Persian",
    "ff": "Fulah",
    "fi": "Finnish",
    "fj": "Fijian",
    "fo": "Faroese",
    "fr": "French",
    "fy": "Western Frisian",
    "ga": "Irish",
    "gd": "Scottish Gaelic, Gaelic",
    "gl": "Galician",
    "gn": "Guarani",
    "gu": "Gujarati",
    "gv": "Manx",
    "ha": "Hausa",
    "he": "Hebrew",
    "hi": "Hindi",
    "ho": "Hiri Motu",
    "hr": "Croatian",
    "ht": "Haitian Creole, Haitian",
    "hu": "Hungarian",
    "hy": "Armenian",
    "hz": "Herero",
    "ia": "Interlingua (International Auxiliary Language Association)",
    "id": "Indonesian",
    "ie": "Interlingue, Occidental",
    "ig": "Igbo",
    "ii": "Sichuan Yi, Nuosu",
    "ik": "Inupiaq",
    "io": "Ido",
    "is": "Icelandic",
    "it": "Italian",
    "iu": "Inuktitut",
    "ja": "Japanese",
    "jv": "Javanese",
    "ka": "Georgian",
    "kg": "Kongo",
    "ki": "Kikuyu, Gikuyu",
    "kj": "Kuanyama, Kwanyama",
    "kk": "Kazakh",
    "kl": "Greenlandic, Kalaallisut",
    "km": "Central Khmer",
    "kn": "Kannada",
    "ko": "Korean",
    "kr": "Kanuri",
    "ks": "Kashmiri",
    "ku": "Kurdish",
    "kv": "Komi",
    "kw": "Cornish",
    "ky": "Kirghiz, Kyrgyz",
    "la": "Latin",
    "lb": "Luxembourgish, Letzeburgesch",
    "lg": "Luganda",
    "li": "Limburgish, Limburger, Limburgan",
    "ln": "Lingala",
    "lo": "Lao",
    "lt": "Lithuanian",
    "lu": "Luba-Katanga",
    "lv": "Latvian",
    "mg": "Malagasy",
    "mh": "Marshallese",
    "mi": "M\u0100ori",
    "mk": "Macedonian",
    "ml": "Malayalam",
    "mn": "Mongolian",
    "mr": "Marathi",
    "ms": "Malay",
    "mt": "Maltese",
    "my": "Burmese",
    "na": "Nauruan",
    "nb": u"Norwegian",
    "nb_NO": "Norwegian Bokm\u00e5l",
    "nd": "Northern Ndebele",
    "ne": "Nepali",
    "ng": "Ndonga",
    "nl": "Dutch",
    "nn": "Norwegian Nynorsk",
    "no": "Norwegian",
    "nr": "Southern Ndebele",
    "nv": "Navajo, Navaho",
    "ny": "Chichewa, Chewa, Nyanja",
    "oc": "Occitan (1500-)",
    "oj": "Ojibwa",
    "om": "Oromo",
    "or": "Oriya",
    "os": "Ossetian, Ossetic",
    "pa": "Punjabi, Panjabi",
    "pi": "Pali",
    "pl": "Polish",
    "ps": "Pashto language, Pashto",
    "pt": "Portuguese",
    "pt_BR": "Portuguese (Brazil)",
    "qu": "Quechua",
    "rm": "Romansh",
    "rn": "Rundi",
    "ro": "Romanian",
    "ru": "Russian",
    "rw": "Kinyarwanda",
    "sa": "Sanskrit",
    "sc": "Sardinian",
    "sd": "Sindhi",
    "se": "Northern Sami",
    "sg": "Sango",
    "si": "Sinhalese, Sinhala",
    "sk": "Slovak",
    "sl": "Slovenian",
    "sm": "Samoan",
    "sn": "Shona",
    "so": "Somali",
    "sq": "Albanian",
    "sr": "Serbian",
    "ss": "Swati",
    "st": "Southern Sotho",
    "su": "Sundanese",
    "sv": "Swedish",
    "sw": "Swahili",
    "ta": "Tamil",
    "te": "Telugu",
    "tg": "Tajik",
    "th": "Thai",
    "ti": "Tigrinya",
    "tk": "Turkmen",
    "tl": "Tagalog",
    "tn": "Tswana",
    "to": "Tonga (Tonga Islands)",
    "tr": "Turkish",
    "ts": "Tsonga",
    "tt": "Tatar",
    "tw": "Twi",
    "ty": "Tahitian",
    "ug": "Uighur, Uyghur",
    "uk": "Ukrainian",
    "ur": "Urdu",
    "uz": "Uzbek",
    "ve": "Venda",
    "vi": "Vietnamese",
    "vo": "Volap\u00FCck",
    "wa": "Walloon",
    "wo": "Wolof",
    "xh": "Xhosa",
    "yi": "Yiddish",
    "yo": "Yoruba",
    "za": "Zhuang, Chuang",
    "zh": "Chinese",
    "zh_CN": "Chinese (Simplified)",
    "zh_HK": "Chinese (Hong Kong)",
    "zh_SG": "Chinese (Singapore)",
    "zh_TW": "Chinese (Taiwan)",
    "zu": "Zulu"
}


def iso6931_code_for_language_name(language):
    return dict((v,k) for k, v in \
        language_name_for_iso6931_code.items())[language]



#
# prefill_tag_behaviour_plugin.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.plugin import Plugin


class PrefillTagBehaviourPlugin(Plugin):

    name = _("Prefill tag behaviour")
    description = \
_("""When creating new cards, Mnemosyne normally prefills the 'Tag(s)' field with the last value you used, regardless of card type.\n
With this plugin, Mnemosyne will for each card type remember the last tag you used, such that e.g. Chinese cards get a prefilled tag like "My Chinese textbook::Chapter 10", whereas Front-to-Back cards get "European Capitals".""")
    supported_API_level = 3

    def activate(self):
        Plugin.activate(self)
        self.config()["is_last_used_tags_per_card_type"] = True

    def deactivate(self):
        Plugin.deactivate(self)
        self.config()["is_last_used_tags_per_card_type"] = False



#
# qpushbutton2.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtGui, QtWidgets


class QPushButton2(QtWidgets.QPushButton):

    """QPushButton which throws away key repeats."""

    def keyPressEvent(self, event):
        if not event.isAutoRepeat():
            return QtWidgets.QPushButton.keyPressEvent(self, event)



#
# libmnemosyne <Peter.Bienstman@UGent.be>
#

import os
import sys
import importlib
import traceback
import warnings
from mnemosyne.libmnemosyne.component import Component
from mnemosyne.libmnemosyne.utils import expand_path, contract_path, \
    traceback_string
from mnemosyne.libmnemosyne.component_manager import new_component_manager, \
    register_component_manager, unregister_component_manager


class Mnemosyne(Component):

    """This class groups the functionality needed to initialise and finalise
    Mnemosyne in a typical scenario.

    See also 'how to write a new frontend' in the docs for more information
    on startup and configuration of libmnemosyne.

    """

    def __init__(self, upload_science_logs, interested_in_old_reps,
        asynchronous_database=False):

        """When 'upload_science_logs' is set to 'None', it means that its
        value is user-specified through the GUI. Explicitly setting this to
        True or False overrides the user choice.

        For mobile clients, it is recommended that you set
        'upload_science_logs' to 'False'. We need to specify this as an
        argument here, so that we can inject it on time to prevent the
        uploader thread from starting.

        'interested_in_old_reps' can be set to 'False' on a mobile client
        which does not show historical statistical data, in order to speed
        up the initial sync and save disk space. We've specified this as a
        non-default argument here, in order to force front-end writers to
        consider it.

        Setting 'asynchronous_database' to True increases the risk of data
        loss and should only be done to speed up the test suite.

        """

        sys.excepthook = self.handle_exception
        self.upload_science_logs = upload_science_logs
        self.interested_in_old_reps = interested_in_old_reps
        self.asynchronous_database = asynchronous_database
        self.component_manager = new_component_manager()
        self.components = [
         ("mnemosyne.libmnemosyne.databases.SQLite",
          "SQLite"),
         ("mnemosyne.libmnemosyne.database",
          "DatabaseMaintenance"),
         ("mnemosyne.libmnemosyne.configuration",
          "Configuration"),
         ("mnemosyne.libmnemosyne.loggers.database_logger",
          "DatabaseLogger"),
         ("mnemosyne.libmnemosyne.stopwatch",
          "Stopwatch"),
         ("mnemosyne.libmnemosyne.card_types.front_to_back",
          "FrontToBack"),
         ("mnemosyne.libmnemosyne.card_types.both_ways",
          "BothWays"),
         ("mnemosyne.libmnemosyne.card_types.vocabulary",
          "Vocabulary"),
         ("mnemosyne.libmnemosyne.card_types.both_ways",
          "FrontToBackToBothWays"),
         ("mnemosyne.libmnemosyne.card_types.both_ways",
          "BothWaysToFrontToBack"),
         ("mnemosyne.libmnemosyne.card_types.vocabulary",
          "FrontToBackToVocabulary"),
         ("mnemosyne.libmnemosyne.card_types.vocabulary",
          "BothWaysToVocabulary"),
         ("mnemosyne.libmnemosyne.card_types.vocabulary",
          "VocabularyToFrontToBack"),
         ("mnemosyne.libmnemosyne.card_types.vocabulary",
          "VocabularyToBothWays"),
         ("mnemosyne.libmnemosyne.render_chains.default_render_chain",
          "DefaultRenderChain"),
         ("mnemosyne.libmnemosyne.render_chains.plain_text_chain",
          "PlainTextChain"),
         ("mnemosyne.libmnemosyne.render_chains.sync_to_card_only_client",
          "SyncToCardOnlyClient"),
         ("mnemosyne.libmnemosyne.render_chains.card_browser_render_chain",
          "CardBrowserRenderChain"),
         ("mnemosyne.libmnemosyne.filters.latex",
          "CheckForUpdatedLatexFiles"),
         ("mnemosyne.libmnemosyne.filters.latex",
          "LatexFilenamesFromData"),
         ("mnemosyne.libmnemosyne.filters.latex",
          "DeleteUnusedLatexFiles"),
         ("mnemosyne.libmnemosyne.filters.latex",
          "PreprocessClozeLatex"),
         ("mnemosyne.libmnemosyne.filters.latex",
          "PostprocessQAClozeLatex"),
         ("mnemosyne.libmnemosyne.controllers.default_controller",
          "DefaultController"),
         ("mnemosyne.libmnemosyne.study_modes.scheduled_forgotten_new",
          "ScheduledForgottenNew"),
         ("mnemosyne.libmnemosyne.study_modes.new_only",
          "NewOnly"),
         ("mnemosyne.libmnemosyne.study_modes.cram_all",
          "CramAll"),
         ("mnemosyne.libmnemosyne.study_modes.cram_recent",
          "CramRecent"),
         ("mnemosyne.libmnemosyne.card_types.map",
          "MapPlugin"),
         ("mnemosyne.libmnemosyne.card_types.cloze",
          "ClozePlugin"),
         ("mnemosyne.libmnemosyne.card_types.sentence",
          "SentencePlugin"),
         ("mnemosyne.libmnemosyne.card_types.M_sided",
          "MSided"),
         ("mnemosyne.libmnemosyne.criteria.default_criterion",
          "DefaultCriterion"),
         ("mnemosyne.libmnemosyne.databases.SQLite_criterion_applier",
          "DefaultCriterionApplier"),
         ("mnemosyne.libmnemosyne.statistics_pages.schedule",
          "Schedule"),
         ("mnemosyne.libmnemosyne.statistics_pages.retention_score",
          "RetentionScore"),
         ("mnemosyne.libmnemosyne.statistics_pages.cards_added",
          "CardsAdded"),
         ("mnemosyne.libmnemosyne.statistics_pages.cards_learned",
          "CardsLearned"),
         ("mnemosyne.libmnemosyne.statistics_pages.grades",
          "Grades"),
         ("mnemosyne.libmnemosyne.statistics_pages.easiness",
          "Easiness"),
         ("mnemosyne.libmnemosyne.statistics_pages.current_card",
          "CurrentCard"),
         ("mnemosyne.libmnemosyne.file_formats.mnemosyne1_mem",
          "Mnemosyne1Mem"),
         ("mnemosyne.libmnemosyne.file_formats.mnemosyne1_xml",
          "Mnemosyne1XML"),
         ("mnemosyne.libmnemosyne.file_formats.mnemosyne2_cards",
          "Mnemosyne2Cards"),
         ("mnemosyne.libmnemosyne.file_formats.mnemosyne2_db",
          "Mnemosyne2Db"),
         ("mnemosyne.libmnemosyne.file_formats.tsv",
          "Tsv"),
         ("mnemosyne.libmnemosyne.file_formats.anki2",
          "Anki2"),
         ("mnemosyne.libmnemosyne.file_formats.supermemo_7_txt",
          "SuperMemo7Txt"),
         ("mnemosyne.libmnemosyne.file_formats.smconv_XML",
          "Smconv_XML"),
         ("mnemosyne.libmnemosyne.file_formats.cuecard_wcu",
          "CuecardWcu"),
         ("mnemosyne.libmnemosyne.translator",
          "Translator"),
         ("mnemosyne.libmnemosyne.pronouncer",
          "Pronouncer"),
         ("mnemosyne.libmnemosyne.languages.afrikaans",
          "Afrikaans"),
         ("mnemosyne.libmnemosyne.languages.albanian",
          "Albanian"),
         ("mnemosyne.libmnemosyne.languages.amharic",
          "Amharic"),
         ("mnemosyne.libmnemosyne.languages.arabic",
          "Arabic"),
         ("mnemosyne.libmnemosyne.languages.armenian",
          "Armenian"),
         ("mnemosyne.libmnemosyne.languages.azerbaijani",
          "Azerbaijani"),
         ("mnemosyne.libmnemosyne.languages.basque",
          "Basque"),
         ("mnemosyne.libmnemosyne.languages.belarusian",
          "Belarusian"),
         ("mnemosyne.libmnemosyne.languages.bengali",
          "Bengali"),
         ("mnemosyne.libmnemosyne.languages.bosnian",
          "Bosnian"),
         ("mnemosyne.libmnemosyne.languages.bulgarian",
          "Bulgarian"),
         ("mnemosyne.libmnemosyne.languages.catalan",
          "Catalan"),
         ("mnemosyne.libmnemosyne.languages.cebuano",
          "Cebuano"),
         ("mnemosyne.libmnemosyne.languages.chinese",
          "Chinese"),
         ("mnemosyne.libmnemosyne.languages.corsican",
          "Corsican"),
         ("mnemosyne.libmnemosyne.languages.croatian",
          "Croatian"),
         ("mnemosyne.libmnemosyne.languages.czech",
          "Czech"),
         ("mnemosyne.libmnemosyne.languages.danish",
          "Danish"),
         ("mnemosyne.libmnemosyne.languages.dutch",
          "Dutch"),
         ("mnemosyne.libmnemosyne.languages.english",
          "English"),
         ("mnemosyne.libmnemosyne.languages.esperanto",
          "Esperanto"),
         ("mnemosyne.libmnemosyne.languages.estonian",
          "Estonian"),
         ("mnemosyne.libmnemosyne.languages.finnish",
          "Finnish"),
         ("mnemosyne.libmnemosyne.languages.french",
          "French"),
         ("mnemosyne.libmnemosyne.languages.frisian",
          "Frisian"),
         ("mnemosyne.libmnemosyne.languages.gaelic",
          "Gaelic"),
         ("mnemosyne.libmnemosyne.languages.galician",
          "Galician"),
         ("mnemosyne.libmnemosyne.languages.georgian",
          "Georgian"),
         ("mnemosyne.libmnemosyne.languages.german",
          "German"),
         ("mnemosyne.libmnemosyne.languages.greek",
          "Greek"),
         ("mnemosyne.libmnemosyne.languages.gujarati",
          "Gujarati"),
         ("mnemosyne.libmnemosyne.languages.haitian",
          "Haitian"),
         ("mnemosyne.libmnemosyne.languages.hausa",
          "Hausa"),
         ("mnemosyne.libmnemosyne.languages.hawaiian",
          "Hawaiian"),
         ("mnemosyne.libmnemosyne.languages.hebrew",
          "Hebrew"),
         ("mnemosyne.libmnemosyne.languages.hindi",
          "Hindi"),
         ("mnemosyne.libmnemosyne.languages.hmong",
          "Hmong"),
         ("mnemosyne.libmnemosyne.languages.hungarian",
          "Hungarian"),
         ("mnemosyne.libmnemosyne.languages.icelandic",
          "Icelandic"),
         ("mnemosyne.libmnemosyne.languages.igbo",
          "Igbo"),
         ("mnemosyne.libmnemosyne.languages.indonesian",
          "Indonesian"),
         ("mnemosyne.libmnemosyne.languages.irish",
          "Irish"),
         ("mnemosyne.libmnemosyne.languages.italian",
          "Italian"),
         ("mnemosyne.libmnemosyne.languages.japanese",
          "Japanese"),
         ("mnemosyne.libmnemosyne.languages.javanese",
          "Javanese"),
         ("mnemosyne.libmnemosyne.languages.kannada",
          "Kannada"),
         ("mnemosyne.libmnemosyne.languages.kazakh",
          "Kazakh"),
         ("mnemosyne.libmnemosyne.languages.khmer",
          "Khmer"),
         ("mnemosyne.libmnemosyne.languages.korean",
          "Korean"),
         ("mnemosyne.libmnemosyne.languages.kurdish",
          "Kurdish"),
         ("mnemosyne.libmnemosyne.languages.kyrgyz",
          "Kyrgyz"),
         ("mnemosyne.libmnemosyne.languages.lao",
          "Lao"),
         ("mnemosyne.libmnemosyne.languages.latin",
          "Latin"),
         ("mnemosyne.libmnemosyne.languages.latvian",
          "Latvian"),
         ("mnemosyne.libmnemosyne.languages.lithuanian",
          "Lithuanian"),
         ("mnemosyne.libmnemosyne.languages.luxembourgish",
          "Luxembourgish"),
         ("mnemosyne.libmnemosyne.languages.macedonian",
          "Macedonian"),
         ("mnemosyne.libmnemosyne.languages.malagasy",
          "Malagasy"),
         ("mnemosyne.libmnemosyne.languages.malay",
          "Malay"),
         ("mnemosyne.libmnemosyne.languages.malayalam",
          "Malayalam"),
         ("mnemosyne.libmnemosyne.languages.maltese",
          "Maltese"),
         ("mnemosyne.libmnemosyne.languages.maori",
          "Maori"),
         ("mnemosyne.libmnemosyne.languages.marathi",
          "Marathi"),
         ("mnemosyne.libmnemosyne.languages.mongolian",
          "Mongolian"),
         ("mnemosyne.libmnemosyne.languages.myanmar",
          "Myanmar"),
         ("mnemosyne.libmnemosyne.languages.nepali",
          "Nepali"),
         ("mnemosyne.libmnemosyne.languages.norwegian",
          "Norwegian"),
         ("mnemosyne.libmnemosyne.languages.nyanja",
          "Nyanja"),
         ("mnemosyne.libmnemosyne.languages.pashto",
          "Pashto"),
         ("mnemosyne.libmnemosyne.languages.persian",
          "Persian"),
         ("mnemosyne.libmnemosyne.languages.polish",
          "Polish"),
         ("mnemosyne.libmnemosyne.languages.portuguese",
          "Portuguese"),
         ("mnemosyne.libmnemosyne.languages.punjabi",
          "Punjabi"),
         ("mnemosyne.libmnemosyne.languages.romanian",
          "Romanian"),
         ("mnemosyne.libmnemosyne.languages.russian",
          "Russian"),
         ("mnemosyne.libmnemosyne.languages.samoan",
          "Samoan"),
         ("mnemosyne.libmnemosyne.languages.serbian",
          "Serbian"),
         ("mnemosyne.libmnemosyne.languages.sesotho",
          "Sesotho"),
         ("mnemosyne.libmnemosyne.languages.shona",
          "Shona"),
         ("mnemosyne.libmnemosyne.languages.sindhi",
          "Sindhi"),
         ("mnemosyne.libmnemosyne.languages.sinhala",
          "Sinhala"),
         ("mnemosyne.libmnemosyne.languages.slovak",
          "Slovak"),
         ("mnemosyne.libmnemosyne.languages.slovenian",
          "Slovenian"),
         ("mnemosyne.libmnemosyne.languages.somali",
          "Somali"),
         ("mnemosyne.libmnemosyne.languages.spanish",
          "Spanish"),
         ("mnemosyne.libmnemosyne.languages.sundanese",
          "Sundanese"),
         ("mnemosyne.libmnemosyne.languages.swahili",
          "Swahili"),
         ("mnemosyne.libmnemosyne.languages.swedish",
          "Swedish"),
         ("mnemosyne.libmnemosyne.languages.tagalog",
          "Tagalog"),
         ("mnemosyne.libmnemosyne.languages.tajik",
          "Tajik"),
         ("mnemosyne.libmnemosyne.languages.tamil",
          "Tamil"),
         ("mnemosyne.libmnemosyne.languages.telugu",
          "Telugu"),
         ("mnemosyne.libmnemosyne.languages.thai",
          "Thai"),
         ("mnemosyne.libmnemosyne.languages.turkish",
          "Turkish"),
         ("mnemosyne.libmnemosyne.languages.ukrainian",
          "Ukrainian"),
         ("mnemosyne.libmnemosyne.languages.urdu",
          "Urdu"),
         ("mnemosyne.libmnemosyne.languages.uzbek",
          "Uzbek"),
         ("mnemosyne.libmnemosyne.languages.vietnamese",
          "Vietnamese"),
         ("mnemosyne.libmnemosyne.languages.welsh",
          "Welsh"),
         ("mnemosyne.libmnemosyne.languages.xhosa",
          "Xhosa"),
         ("mnemosyne.libmnemosyne.languages.yiddish",
          "Yiddish"),
         ("mnemosyne.libmnemosyne.languages.yoruba",
          "Yoruba"),
         ("mnemosyne.libmnemosyne.languages.zulu",
          "Zulu")]
        try: # On Android, importlib.util doesn't get found, even though it's present...
            if (importlib.util.find_spec("gtts")) is not None:
                self.components.append(\
                    ("mnemosyne.libmnemosyne.pronouncers.google_pronouncer",
                    "GooglePronouncer"))
            else:
                warnings.warn("gTTS package is not installed. Text to speak feature will not be available.")
            if ((importlib.util.find_spec("googletrans")) is not None):
                self.components.append(\
                    ("mnemosyne.libmnemosyne.translators.google_translator", "GoogleTranslator"))
            else:
                warnings.warn("google_trans package is not installed. Google translate feature will not be available.")
        except:
            pass
        self.gui_for_component = {}

    def handle_exception(self, type, value, tb):
        body = "An unexpected error has occurred.\n" + \
            "Please forward the following info to the developers:\n\n" + \
            "Traceback (innermost last):\n"
        list = traceback.format_tb(tb, limit=None) + \
               traceback.format_exception_only(type, value)
        body = body + "%-20s %s" % ("".join(list[:-1]), list[-1])
        print("Log body:\n", body)
        try:
            if sys.platform != "win32":
                sys.stderr.write(body)
            if type is KeyboardInterrupt:
                self.main_widget().handle_keyboard_interrupt(body)
            else:
                self.main_widget().show_error(body)
        except:
            sys.stderr.write(body)

    def initialise(self, data_dir=None, config_dir=None,
                   filename=None, automatic_upgrades=True, debug_file=None,
                   server_only=False):

        """The automatic upgrades of the database can be turned off by setting
        'automatic_upgrade' to False. This is mainly useful for the testsuite.

        """

        if debug_file:
            self.component_manager.debug_file = open(debug_file, "w")
        self.register_components()
        # Upgrade from 1.x if needed.
        if automatic_upgrades:
            from mnemosyne.libmnemosyne.upgrades.upgrade1 import Upgrade1
            Upgrade1(self.component_manager).backup_old_dir()
        if data_dir:
            self.config().data_dir = data_dir
            self.config().config_dir = data_dir
        if config_dir:
            self.config().config_dir = config_dir
        self.activate_components()
        register_component_manager(self.component_manager,
                                   self.config()["user_id"])
        self.execute_user_plugin_dir()
        self.activate_saved_plugins()
        # If we are only running a sync or a review server, do not yet load
        # the database to prevent threading access issues.
        if server_only:
            if filename:
                self.config()["last_database"] = \
                    contract_path(filename, self.config().data_dir)
            return
        # Loading the database should come after all user plugins have been
        # loaded, since these could be needed e.g. for a card type in the
        # database.
        if filename and not filename.endswith(".db"):
            from mnemosyne.libmnemosyne.gui_translator import _
            self.main_widget().show_error(\
                _("Command line argument is not a *.db file."))
            sys.exit()
        self.load_database(filename)
        # Only now that the database is loaded, we can start writing log
        # events to it. This is why we log started_scheduler and
        # loaded_database manually.
        try:
            self.log().started_program()
        except Exception as e:
            if "lock" in str(e):
                from mnemosyne.libmnemosyne.gui_translator import _
                self.main_widget().show_error(\
                 _("Another copy of Mnemosyne is still running.") + "\n" + \
                 _("Continuing is impossible and will lead to data loss!"))
                sys.exit()
            else:
                raise e
        # Upgrade from 1.x if needed.
        if automatic_upgrades:
            from mnemosyne.libmnemosyne.upgrades.upgrade1 import Upgrade1
            Upgrade1(self.component_manager).run()
        # Finally, we can start the main widget and the review.
        self.main_widget().activate()
        self.start_review()

    def register_components(self):

        """We register all components, but don't activate them yet, because in
        order to activate certain components, certain other components already
        need to be registered. Also, the activation needs to happen in a
        predefined order, and we don't want to burden UI writers with listing
        the components in the correct order.

        """

        for module_name, class_name in self.components:
            component = getattr(\
                importlib.import_module(module_name), class_name)
            if component.instantiate == Component.IMMEDIATELY:
                component = component(component_manager=self.component_manager)
            self.component_manager.register(component)
        for component_name in self.gui_for_component:
            for gui_module_name, gui_class_name in \
                    self.gui_for_component[component_name]:
                gui_class = getattr(\
                    importlib.import_module(gui_module_name), gui_class_name)
                self.component_manager.add_gui_to_component(\
                    component_name, gui_class)

    def activate_components(self):

        """Now that everything is registered, we can activate the components
        in the correct order: first config, followed by log and then the rest.

        """

        # Activate config and inject necessary settings.
        try:
            self.component_manager.current("config").activate()
        except RuntimeError as e:
            self.main_widget().show_error(str(e))
        # Allow front end programmer to override the user setting.
        if self.upload_science_logs is False:
            self.config()["upload_science_logs"] = False
        if self.upload_science_logs is True:
            self.config()["upload_science_logs"] = True
        self.config()["interested_in_old_reps"] = self.interested_in_old_reps
        self.config()["asynchronous_database"] = self.asynchronous_database
        # Activate other components.
        for component in ["log", "gui_translator", "database", "controller"]:
            try:
                self.component_manager.current(component).activate()
            except RuntimeError as e:
                self.main_widget().show_error(str(e))
        sync_server = self.component_manager.current("sync_server")
        if sync_server:
            sync_server.activate()
        web_server = self.component_manager.current("web_server")
        if web_server:
            web_server.activate()

    def execute_user_plugin_dir(self):
        # Note that we put user plugins in the data_dir and not the
        # config_dir as there could be plugins (e.g. new card types) for
        # which the database does not make sense without them.
        plugin_dir = os.path.join(self.config().data_dir, "plugins")
        if plugin_dir not in sys.path:
            sys.path.insert(0, plugin_dir)
        for component in os.listdir(plugin_dir):
            if component.endswith(".py"):
                try:
                    __import__(component[:-3])
                except:
                    from mnemosyne.libmnemosyne.gui_translator import _
                    msg = _("Error when running plugin:") \
                          + "\n" + traceback_string()
                    self.main_widget().show_error(msg)

    def activate_saved_plugins(self):
        for plugin in self.config()["active_plugins"]:
            try:
                for p in self.plugins():
                    if plugin == p.__class__.__name__:
                        p.activate()
                        break
            except:
                from mnemosyne.libmnemosyne.gui_translator import _
                msg = _("Error when running plugin:") \
                      + "\n" + traceback_string()
                self.main_widget().show_error(msg)

    def load_database(self, filename):
        if not filename:
            filename = self.config()["last_database"]
        path = expand_path(filename, self.config().data_dir)
        try:
            if not os.path.exists(path):
                try:
                    self.database().new(path)
                except Exception as e:
                    from mnemosyne.libmnemosyne.gui_translator import _
                    raise RuntimeError(\
                        _("Previous drive letter no longer available."))
            else:
                self.database().load(path)
            self.controller().update_title()
        except RuntimeError as e:
            self.main_widget().show_error(str(e))
            from mnemosyne.libmnemosyne.gui_translator import _
            self.main_widget().show_information(\
_("If you are using a USB key, refer to the instructions on the website so as not to be affected by drive letter changes."))
            # Try to open a new database, but not indefinitely, otherwise this
            # could turn a crash into a much nastier one on Android.
            success = False
            counter = 0
            while not success and counter <= 5:
                counter += 1
                try:
                    self.database().abandon()
                    self.controller().show_open_file_dialog()
                    success = True
                except RuntimeError as e:
                    self.main_widget().show_error(str(e))

    def start_review(self):
        self.controller().set_study_mode(self.study_mode_with_id(\
            self.config()["study_mode"]))
        # Design wart: log().load_database() needs the scheduler to be active,
        # so we can only call it here.
        self.log().loaded_database()
        # No need to log the future schedule everytime we change study
        # mode, so we only do it only on program start and roll-over.
        self.log().future_schedule()

    def finalise(self):
        # Deactivate the sync server first, so that we make sure it reverts
        # to the right backup file in case of problems.
        server = self.component_manager.current("sync_server")
        if server:
            server.deactivate()
        # Ditto for the web server.
        server = self.component_manager.current("web_server")
        if server:
            server.deactivate()
        # Saving the config should happen before we deactivate the plugins,
        # otherwise they are not restored upon reload.
        self.config().save()
        user_id = self.config()["user_id"]
        # We need to log before we unload the database.
        self.log().saved_database()
        self.log().stopped_program()
        # Now deactivate the database, such that deactivating plugins with
        # card types does not raise an error about card types in use.
        self.database().deactivate()
        self.component_manager.unregister(self.database())
        # Then do the review widget and other components.
        if self.review_widget():
            self.review_widget().deactivate()
        self.component_manager.deactivate_all()
        unregister_component_manager(user_id)
        if self.component_manager.debug_file:
            self.component_manager.debug_file.close()



#
# escape_to_html_for_card_browser.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.filter import Filter


class EscapeToHtmlForCardBrowser(Filter):

    """Make sure tags like img, latex, ... show up as tags."""

    def run(self, text, card, fact_key, **render_args):
        if text:
            text = text.replace("<img", "&lt;img")
            text = text.replace("<latex", "&lt;latex")
            text = text.replace("<audio", "&lt;audio")
            text = text.replace("<video", "&lt;video")
        return text



#
# gettext_gui_translator.py <Johannes.Baiter@gmail.com>
#

from mnemosyne.libmnemosyne.gui_translator import GuiTranslator

import os
import sys
import gettext


class GetTextGuiTranslator(GuiTranslator):

    def __init__(self, component_manager):
        GuiTranslator.__init__(self, component_manager)
        # Check if we're running in a development environment.
        if os.path.exists("mo"):
            self.lang_path = "mo"
        else:  # pragma: no cover
            self.lang_path = os.path.join(sys.exec_prefix, "share", "locale")

    def supported_languages(self):
        try:
            return [os.path.split(x)[1] for x in os.listdir(self.lang_path) \
            if os.path.isdir(os.path.join(self.lang_path, x)) and \
            os.path.exists(os.path.join(self.lang_path, x, "LC_MESSAGES")) \
            and "mnemosyne.mo" in os.listdir(\
            os.path.join(self.lang_path, x, "LC_MESSAGES"))]
        except:
            return []

    def set_translator(self, language):
        self._translator = gettext.translation("mnemosyne",
            localedir=self.lang_path, languages=[language],
            fallback=True).gettext




#
# shortcuts.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore

from mnemosyne.libmnemosyne.plugin import Plugin
from mnemosyne.pyqt_ui.review_wdgt import ReviewWdgt


class MyReviewWdgt(ReviewWdgt):

    def __init__(self, **kwds):
        super().__init__(**kwds)
        # Map Q to grade 0
        self.key_to_grade_map[QtCore.Qt.Key_Q] = 0


class ShortcutsPlugin(Plugin):

    name = "Custom shortcuts"
    description = "Customise review widget shortcuts."
    gui_for_component = {"ScheduledForgottenNew" :
        [("shortcuts", "MyReviewWdgt")]}
    gui_for_component = {"NewOnly" :
        [("shortcuts", "MyReviewWdgt")]}
    gui_for_component = {"CramAll" :
        [("shortcuts", "MyReviewWdgt")]}
    gui_for_component = {"CramRecent" :
        [("shortcuts", "MyReviewWdgt")]}
    supported_API_level = 3


# Register plugin.

from mnemosyne.libmnemosyne.plugin import register_user_plugin
plugin = register_user_plugin(ShortcutsPlugin)



#
# google_translator.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.translator import Translator

#from google_trans_new import google_translator
from googletrans import Translator as google_translator
translator = None

class GoogleTranslator(Translator):

    # https://cloud.google.com/translate/docs/languages

    used_for = "af", "sq", "am", "ar", "hy", "az", "eu", "be", "bn", "bs", \
        "bg", "ca", "ceb", "zh", "zh-CN", "zh-TW", "co", "hr", "cs", "da", \
        "nl", "en", "eo", "et", "fi", "fr", "fy", "gl", "ka", "de", "el", \
        "gu", "ht", "ha", "haw", "he", "hi", "hmn", "hu", "is", "ig", "id", \
        "ga", "it", "ja", "jw", "kn", "kk", "km", "ko", "ku", "ky", "lo", \
        "la", "lv", "lt", "lb", "mk", "mg", "ms", "ml", "mt", "mi", "mr", \
        "mn", "my", "ne", "no", "ny", "ps", "fa", "pl", "pt", "pa", "ro", \
        "ru", "sm", "gd", "sr", "st", "sn", "sd", "si", "sk", "sl", "so", \
        "es", "su", "sw", "sv", "tl", "tg", "ta", "te", "th", "tr", "uk", \
        "ur", "uz", "vi", "cy", "xh", "yi", "yo", "zu"

    popup_menu_text = "Insert Google translation..."

    def translate(self, card_type, foreign_text, target_language_id,
                  retries=0):
        global translator
        if not translator:
            translator = google_translator()
        source_language_id = self.config().card_type_property(\
            "language_id", card_type)
        try:
            #result = translator.translate(foreign_text,
            #    lang_tgt=target_language_id, lang_src=source_language_id).strip()
            result = translator.translate(foreign_text,
                dest=target_language_id, src=source_language_id).text.strip()
        except:
            if retries < 5:
                import time; time.sleep(2)
                translator = google_translator()
                return self.translate(card_type, foreign_text,
                    target_language_id, retries=retries+1)
            else:
                return None
        return result



#
# file_format.py <Peter.Bienstman@UGent.be>
#

import os

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.component import Component


class FileFormat(Component):

    component_type = "file_format"
    description = ""
    extension = "" # E.g. ".xml"
    filename_filter = ""  # E.g. "XML Files (*.xml *XML)"
    import_possible = False
    export_possible = False

    def do_import(self, filename, extra_tag_name=None):

        """Make sure fileformats call this implementation first."""

        self.import_dir = os.path.dirname(os.path.abspath(filename))




#
# cards_added.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.statistics_page import PlotStatisticsPage


class CardsAdded(PlotStatisticsPage):

    name = _("Cards added")

    LAST_WEEK = 1
    LAST_MONTH = 2
    LAST_3_MONTHS = 3
    LAST_6_MONTHS = 4
    LAST_YEAR = 5

    variants = [(LAST_WEEK, _("Last week")),
                (LAST_MONTH, _("Last month")),
                (LAST_3_MONTHS, _("Last 3 months")),
                (LAST_6_MONTHS, _("Last 6 months")),
                (LAST_YEAR, _("Last year"))]

    def prepare_statistics(self, variant):
        if variant == self.LAST_WEEK:
            self.x = list(range(-7, 1, 1))
        elif variant == self.LAST_MONTH:
            self.x = list(range(-31, 1, 1))
        elif variant == self.LAST_3_MONTHS:
            self.x = list(range(-91, 1, 1))
        elif variant == self.LAST_6_MONTHS:
            self.x = list(range(-182, 1, 1))
        elif variant == self.LAST_YEAR:
            self.x = list(range(-365, 1, 1))
        else:
            raise AttributeError("Invalid variant")
        self.main_widget().set_progress_text(_("Calculating statistics..."))
        self.main_widget().set_progress_range(len(self.x))
        self.main_widget().set_progress_update_interval(3)
        self.y = []
        for day in self.x:
            self.y.append(self.database().card_count_added_n_days_ago(n=-day))
            self.main_widget().increase_progress(1)
        self.main_widget().close_progress()




#
# cram_recent.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.study_mode import StudyMode
from mnemosyne.libmnemosyne.schedulers.cramming import Cramming
from mnemosyne.libmnemosyne.review_controllers.SM2_controller_cramming \
     import SM2ControllerCramming


class CramRecent(StudyMode):

    id = "CramRecent"
    name = _("Cram recently learned new cards")
    menu_weight = 4
    Scheduler = Cramming
    ReviewController = SM2ControllerCramming

    def activate(self):
        self.activate_components()
        self.review_controller().reset(new_only=True)


from PyQt5 import QtCore, QtWidgets, QtGui

class TagCompleter(QtWidgets.QCompleter):
    def __init__(self, parent = None):
        super(TagCompleter, self).__init__(parent)
        self.setCaseSensitivity(QtCore.Qt.CaseInsensitive)

    def handle_prefix_change(self, new_prefix):
        if new_prefix:
            self.setCompletionPrefix(new_prefix)
            self.complete()

    def refresh_completion_model(self, new_model):
        """Sets the completion model that the completer will use.

        Ideally, we'd be able to just set the completion model once and then
        forget about it. However, this would result in unwanted behavior
        wherein word prefixes are added to the completion list by Qt.

        Example: user types in 'myf' and then selects the pop-up completion
        of 'myfancytag'. This results in 'myf' getting added as a potential
        completion suggestion, which is obviously wrong (... and infuriating
        to debug).

        That's why we only update the model: (1) at the start, with all
        available tags, and (2) when the user actually creates a new tag.
        """
        initial_tag_model_ = QtGui.QStandardItemModel()
        for index in range(0, new_model.rowCount()):
            initial_tag_model_.appendRow(new_model.item(index).clone())
        self.setModel(initial_tag_model_)




#
# custom_tag.py <Peter.Bienstman@UGent.be>
#

import subprocess

from mnemosyne.libmnemosyne.filter import Filter
from mnemosyne.libmnemosyne.plugin import Plugin


class CustomTag(Filter):

    tag_name = "my_tag"
    tag_program = ["/bin/cat", "--show-ends"]

    """This plugin could run several times when a card is displayed,
    so we need to make sure that we run the program only once per card.

    """

    def __init__(self, component_manager):
        Filter.__init__(self, component_manager)
        self.last_filename = None

    def run(self, text, card, fact_key, **render_args):
        if "no_side_effects" in render_args and \
            render_args["no_side_effects"] == True:
            return text
        i = text.lower().find(self.tag_name + " src")
        while i != -1:
            start = text.find("\"", i)
            end = text.find("\"", start + 1)
            if end == -1:
                return text
            filename = text[start+1:end].replace("file://", "")
            if filename == self.last_filename:
                return text
            try:
                subprocess.call(self.tag_program + [filename])
            except:
                self.main_widget.show_error("Unable to open ", filename)
            self.last_filename = filename
            i = text.lower().find(self.tag_name + " src", i + 1)
        return text


class CustomTagPlugin(Plugin):

    name = "Custom tag"
    description = "Intercepts custom tags like <my_tag src=\"filename\"> and runs them in an external program.\n\nEdit the source to customise."
    components = [CustomTag]
    supported_API_level = 3

    def activate(self):
        Plugin.activate(self)
        self.render_chain("default").\
            register_filter(CustomTag, in_front=False)
        # Other chain you might want to add to is e.g. "card_browser".

    def deactivate(self):
        Plugin.deactivate(self)
        self.render_chain("default").\
            unregister_filter(CustomTag)

# Register plugin.

from mnemosyne.libmnemosyne.plugin import register_user_plugin
register_user_plugin(CustomTagPlugin)




#
# mnemosyne1_mem.py <Peter.Bienstman@UGent.be>
#

import os
import re
import sys
import time
import pickle

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.utils import MnemosyneError
from mnemosyne.libmnemosyne.file_format import FileFormat
from mnemosyne.libmnemosyne.file_formats.mnemosyne1 import Mnemosyne1
from mnemosyne.libmnemosyne.file_formats.science_log_parser \
     import ScienceLogParser


class Mnemosyne1Mem(FileFormat, Mnemosyne1):

    description = _("Mnemosyne 1.x *.mem files")
    extension = ".mem"
    filename_filter = _("Mnemosyne 1.x *.mem databases (*.mem)")
    import_possible = True
    export_possible = False

    def do_import(self, filename, extra_tag_names=""):
        FileFormat.do_import(self, filename, extra_tag_names)
        w = self.main_widget()
        w.set_progress_text(_("Importing cards..."))
        db = self.database()
        # The import process generates card log entries, which we will delete
        # in favour of those events that are recorded in the logs and which
        # capture the true timestamps. They also have new 2.0 ids, as opposed
        # to their old 1.x ids.
        log_index = db.current_log_index()
        try:
            self.read_items_from_mnemosyne1_mem(filename)
            self.create_cards_from_mnemosyne1(extra_tag_names)
        except MnemosyneError:
            w.close_progress()
            return
        db.remove_card_log_entries_since(log_index)
        self.import_logs(filename)
        # Force an ADDED_CARD log entry for those cards that did not figure in
        # the txt logs, e.g. due to missing or corrupt logs.
        db.add_missing_added_card_log_entries(
            set(item.id for item in self.items))
        # In 2.x, repetition events are used to update a card's last_rep and
        # next_rep during sync. In 1.x, there was no such information, and
        # calculating it from the logs will fail if they are incomplete.
        # Therefore, we force a card edit event for all cards.
        timestamp = int(time.time())
        for item in self.items:
            db.log_edited_card(timestamp, item.id)
        # Detect inverses.
        db.link_inverse_cards()
        w.close_progress()
        self.warned_about_missing_media = False

    def read_items_from_mnemosyne1_mem(self, filename):
        sys.modules["mnemosyne.core"] = object()
        sys.modules["mnemosyne.core.mnemosyne_core"] \
            = Mnemosyne1.MnemosyneCore()
        # For importing Python 2 pickles, we run into this bug:
        # http://bugs.python.org/issue22005
        # Workaround is opening this file using 'bytes' encoding, but
        # this requires extra work for us in setting up the data members.
        try:
            memfile = open(filename, "rb")
            header = memfile.readline()
            self.starttime, self.categories, self.items \
                = pickle.load(memfile, encoding="bytes")
            self.starttime = self.starttime.__dict__[b"time"]
            for category in self.categories:
                category.name = category.__dict__[b"name"]
                category.active = category.__dict__[b"active"]
                del category.__dict__[b"name"]
                del category.__dict__[b"active"]
            for item in self.items:
                if type(item.__dict__[b"id"]) == bytes:
                    item.id = str(item.__dict__[b"id"], "utf-8")
                else:
                    item.id = item.__dict__[b"id"]
                item.cat = item.__dict__[b"cat"]
                item.q = item.__dict__[b"q"]
                item.a = item.__dict__[b"a"]
                item.unseen = item.__dict__[b"unseen"]
                item.grade = item.__dict__[b"grade"]
                item.next_rep = item.__dict__[b"next_rep"]
                item.last_rep = item.__dict__[b"last_rep"]
                item.easiness = item.__dict__[b"easiness"]
                item.acq_reps = item.__dict__[b"acq_reps"]
                item.ret_reps = item.__dict__[b"ret_reps"]
                item.lapses = item.__dict__[b"lapses"]
                item.acq_reps_since_lapse = \
                    item.__dict__[b"acq_reps_since_lapse"]
                item.ret_reps_since_lapse = \
                    item.__dict__[b"ret_reps_since_lapse"]
                del item.__dict__[b"id"]
                del item.__dict__[b"cat"]
                del item.__dict__[b"q"]
                del item.__dict__[b"a"]
                del item.__dict__[b"unseen"]
                del item.__dict__[b"grade"]
                del item.__dict__[b"next_rep"]
                del item.__dict__[b"last_rep"]
                del item.__dict__[b"easiness"]
                del item.__dict__[b"acq_reps"]
                del item.__dict__[b"ret_reps"]
                del item.__dict__[b"lapses"]
                del item.__dict__[b"acq_reps_since_lapse"]
                del item.__dict__[b"ret_reps_since_lapse"]
        except (FileNotFoundError, PermissionError) as e:
            self.main_widget().show_error(_("Unable to open file."))
            raise MnemosyneError
        except Exception as e:
            import traceback
            self.main_widget().show_error(traceback.format_exc())
            raise MnemosyneError

    def import_logs(self, filename):
        w = self.main_widget()
        db = self.database()
        w.set_progress_text(_("Importing history..."))
        log_dir = os.path.join(os.path.dirname(filename), "history")
        if not os.path.exists(log_dir):
            w.close_progress()
            w.show_information(_("No history found to import."))
            return
        # The events that we import from the science logs obviously should not
        # be reexported to these logs (this is true for both the archived logs
        # and log.txt). So, before the import, we flush the SQL logs to the
        # science logs, and after the import we edit the partership index to
        # skip these entries.
        db.dump_to_science_log()
        # Manage database indexes.
        db.before_1x_log_import()
        filenames = [os.path.join(log_dir, logname) for logname in \
            sorted(os.listdir(log_dir)) if logname.endswith(".bz2")]
        # log.txt can also contain data we need to import, especially on the
        # initial upgrade from 1.x. 'ids_to_parse' will make sure we only pick
        # up the relevant events. (If we do the importing after having used
        # 2.x for a while, there could be duplicate load events, etc, but these
        # don't matter.)
        filenames.append(os.path.join(os.path.dirname(filename), "log.txt"))
        w.set_progress_range(len(filenames))
        ignored_files = []
        parser = ScienceLogParser(self.database(),
            ids_to_parse=self.items_by_id,
            machine_id=self.config().machine_id())
        for filename in filenames:
            try:
                parser.parse(filename)
            except:
                ignored_files.append(filename)
            w.increase_progress(1)
        if ignored_files:
            w.show_information(_("Ignoring unparsable files:<br/>") +\
                '<br/>'.join(ignored_files))
        # Manage database indexes.
        db.after_1x_log_import()
        db.skip_science_log()



#
# new_card_type.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.plugin import Plugin
from mnemosyne.libmnemosyne.fact_view import FactView
from mnemosyne.libmnemosyne.card_types.vocabulary import Vocabulary

class DecoratedVocabulary(Vocabulary):

    id = "3_decorated"
    name = _("Vocabulary (decorated)")

    # The keys we inherit from Vocabulary, we just override the FactViews.

    # Recognition.
    v1 = FactView(_("Recognition"), "3::1")
    v1.q_fact_keys = ["f"]
    v1.a_fact_keys = ["p_1", "m_1", "n"]
    v1.q_fact_key_decorators = {"f": "What is the translation of ${f}?"}

    # Production.
    v2 = FactView(_("Production"), "3::2")
    v2.q_fact_keys = ["m_1"]
    v2.a_fact_keys = ["f", "p_1", "n"]
    v2.q_fact_key_decorators = {"m_1": "How do you say ${m_1}?"}

    fact_views = [v1, v2]


# Wrap it into a Plugin and then register the Plugin.

class DecoratedVocabularyPlugin(Plugin):

    name = "Decorated vocabulary"
    description = "Vocabulary card type with some extra text"
    components = [DecoratedVocabulary]
    supported_API_level = 3

from mnemosyne.libmnemosyne.plugin import register_user_plugin
register_user_plugin(DecoratedVocabularyPlugin)







#
# ui_component.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.component import Component


class UiComponent(Component):

    """Apart from the main widget, UI components are instantiated late for
    efficiency reasons.

    """

    component_type = "ui_component"

    instantiate = Component.LATER




#
# completion_combo_box.py: Emilian Mihalache <emihalac@gmail.com>
#

from PyQt5 import QtCore, QtGui, QtWidgets
from mnemosyne.pyqt_ui.tag_line_edit import TagLineEdit


class CompletionComboBox(QtWidgets.QComboBox):

    def __init__(self, parent=None):
        super(CompletionComboBox, self).__init__(parent)

        tag_line_edit_ = TagLineEdit(self)
        self.setLineEdit(tag_line_edit_)

        # Disable builtin autocompletion.
        # We'll be replacing it with an autocompletion mechanism
        #  that can work with elements after the first one.
        #
        # This _needs_ to be called _after_ setLineEdit(),
        #  because setLineEdit() will create&store a new QCompleter,
        #  which we're deleting here.
        self.setCompleter(None)

    def refresh_completion_model(self):
        self.lineEdit().refresh_completion_model(self.model())



#
# review_wdgt.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.UDP_server.UDP_widget import UDP_Widget
from mnemosyne.libmnemosyne.ui_components.review_widget import ReviewWidget


class ReviewWdgt(ReviewWidget, UDP_Widget):

    def __init__(self, component_manager):
        ReviewWidget.__init__(self, component_manager)

    def show_answer(self):
        self.callback()

    def grade_answer(self, grade):
        self.callback(grade)

    def set_question_box_visible(self, is_visible):
        self.callback(is_visible)

    def set_answer_box_visible(self, is_visible):
        self.callback(is_visible)

    def set_question_label(self, text):
        self.callback(text)

    def set_question(self, text):
        self.callback(text)

    def set_answer(self, text):
        self.callback(text)

    def clear_question(self):
        self.callback()

    def clear_answer(self):
        self.callback()

    def update_show_button(self, text, is_default, is_enabled):
        self.callback(text, is_default, is_enabled)

    def set_grades_enabled(self, is_enabled):
        self.callback(is_enabled)

    def set_grade_enabled(self, grade, is_enabled):
        self.callback(grade, is_enabled)

    def set_default_grade(self, grade):
        self.callback(grade)

    def set_grades_title(self, text):
        self.callback(text)

    def set_grade_text(self, grade, text):
        self.callback(grade, text)

    def set_grade_tooltip(self, grade, text):
        self.callback(grade, text)

    def update_status_bar_counters(self):
        self.callback()

    def redraw_now(self):
        self.callback()



#
# pie_chart_statistics.py <Peter.Bienstman@UGent.be>, <mike@peacecorps.org.cv>
#

from mnemosyne.libmnemosyne.plugin import Plugin
from mnemosyne.libmnemosyne.statistics_page import PlotStatisticsPage
from mnemosyne.pyqt_ui.statistics_wdgts_plotting import PlotStatisticsWdgt

# The piechart statistics page (GUI independent part).

class MyGrades(PlotStatisticsPage):

    name = "Grades pie chart"

    def prepare_statistics(self, variant):
        self.x = list(range(-1, 6))
        self.y = []
        for grade in self.x:
            self.y.append(self.database().con.execute(\
                "select count() from cards where grade=? and active=1",
                 (grade, )).fetchone()[0])

# The custom widget.

class PieChartWdgt(PlotStatisticsWdgt):

    used_for = MyGrades

    def show_statistics(self, variant):
        if not self.page.y:
            self.display_message(_("No stats available."))
            return
        self.activate()
        # Pie charts look better on a square canvas.
        self.axes.set_aspect('equal')
        labels = ["Unseen" if self.page.y[0] > 0 else ""] + \
            ["Grade %d" % (g-1) if self.page.y[g] > 0 \
            else "" for g in range(1, 7)]
        colors = ["w", "r", "m", "y", "g", "c", "b"]
        # Only print percentage on wedges > 5%.
        autopct = lambda x: "%1.1f%%" % x if x > 5 else ""
        self.axes.pie(self.page.y, labels=labels, colors=colors,
                      shadow=True, autopct=autopct)
        self.axes.set_title("Number of cards")

# Wrap it into a Plugin and then register the Plugin.

class PieChartPlugin(Plugin):
    name = "Pie chart grades"
    description = "Show the grade statistics in a pie chart"
    components = [MyGrades, PieChartWdgt]
    supported_API_level = 3

from mnemosyne.libmnemosyne.plugin import register_user_plugin
register_user_plugin(PieChartPlugin)




#
# test_render_chain.py <Peter.Bienstman@UGent.be>
#

import os

from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne.filter import Filter
from mnemosyne.libmnemosyne.renderer import Renderer


class TestRenderChain(MnemosyneTest):

    def test_add_filter(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
            grade=-1, tag_names=["default"])[0]

        class MyFilter(Filter):
            def run(self, text, card, fact_key):
                return "666"

        self.render_chain().register_filter(MyFilter)

        assert "666" in card.question()
        assert isinstance(self.render_chain().filter(MyFilter), MyFilter)
        assert self.render_chain().filter(TestRenderChain) is None

        assert type(self.render_chain()._filters[0]) \
               != type(MyFilter(self.mnemosyne.component_manager))

        self.render_chain().unregister_filter(MyFilter)
        assert "666" not in card.question()

        self.render_chain().unregister_filter(type(1))

    def test_add_filter_2(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
            grade=-1, tag_names=["default"])[0]

        class MyFilter(Filter):
            def run(self, text, card, fact_key):
                return "666"

        self.render_chain().register_filter(MyFilter, in_front=True)
        assert "666" in card.question()

        assert type(self.render_chain()._filters[0]) \
               == type(MyFilter(self.mnemosyne.component_manager))

    def test_add_filter_order(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
            grade=-1, tag_names=["default"])[0]

        class MyFilter1(Filter):
            def run(self, text, card, fact_key):
                return "666"

        class MyFilter2(Filter):
            def run(self, text, card, fact_key):
                return "[%s]" % text

        class MyFilter3(Filter):
            def run(self, text, card, fact_key):
                return "(%s)" % text

        class MyFilter4(Filter):
            def run(self, text, card, fact_key):
                return "{%s}" % text

        def equals(filter_class):
            ty = type(filter_class(self.mnemosyne.component_manager))
            def eq(x):
                return type(x) == ty
            return eq

        self.render_chain().register_filter_at_front(MyFilter1)
        assert "666" in card.question()

        self.render_chain().register_filter_at_front(MyFilter2, [MyFilter1.__name__])
        assert "[666]" in card.question()

        self.render_chain().register_filter_at_back(MyFilter3, [MyFilter2.__name__])
        assert "[(666)]" in card.question()

        self.render_chain().register_filter_at_front(MyFilter4, [MyFilter3.__name__])
        assert "[{(666)}]" in card.question()

        assert type(self.render_chain()._filters[0]) \
               == type(MyFilter1(self.mnemosyne.component_manager))

        assert type(self.render_chain()._filters[1]) \
               == type(MyFilter3(self.mnemosyne.component_manager))

        assert type(self.render_chain()._filters[2]) \
               == type(MyFilter4(self.mnemosyne.component_manager))

        assert type(self.render_chain()._filters[3]) \
               == type(MyFilter2(self.mnemosyne.component_manager))

    def test_add_card_type_renderer(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
            grade=-1, tag_names=["default"])[0]

        class MyRenderer(Renderer):
            used_for = card_type_1
            def render(self, fact_data, fields, card_type, **render_args):
                return "666"

        self.render_chain().register_renderer(MyRenderer)
        assert "666" in card.question()

        card_type_2 = self.card_type_with_id("2")
        card = self.controller().create_new_cards(fact_data, card_type_2,
            grade=-1, tag_names=["default"])[0]
        assert "666" not in card.question()

        self.render_chain().unregister_renderer(MyRenderer)
        assert "666" not in card.question()

        self.render_chain().unregister_renderer(type(1))

    def test_latex(self):
        fact_data = {"f": "<latex>1<2</latex>",
                     "b": "answer"}
        card_type_1 = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type_1,
            grade=-1, tag_names=["default"])[0]
        card.question()

        filename = os.path.join(os.path.abspath("dot_test"),
            "default.db_media", "_latex", "tmp.tex")
        contents = "".join(open(filename).readlines())
        assert '<' in contents
        assert "&lt;" not in contents



#
# sentence.py <Peter.Bienstman@UGent.be>
#

import re
import copy

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.card import Card
from mnemosyne.libmnemosyne.plugin import Plugin
from mnemosyne.libmnemosyne.card_type import CardType
from mnemosyne.libmnemosyne.fact_view import FactView
from mnemosyne.libmnemosyne.card_types.cloze import Cloze

cloze_re = re.compile(r"\[(.+?)\]", re.DOTALL)


class Sentence(Cloze):

    """A card type using sentences to study foreign languages.

    Apart from simple recognition of the sentence, you can also add production
    cards using close deletion. E.g. if in the sentence field you write
    "La [casa:house] es [grande:big]", you'll get cards with questions like
    "La [house] es grande".

    """

    id = "6"
    name = _("Sentence")

    # List and name the keys.
    fact_keys_and_names = [("f", _("Sentence")),
                           ("p_1", _("Pronunciation")),
                           ("m_1", _("Meaning")),
                           ("n", _("Notes"))]

    # Recognition.
    v1 = FactView(_("Recognition"), "6.1")
    v1.q_fact_keys = ["f"]
    v1.a_fact_keys = ["p_1", "m_1", "n"]

    # Production.
    v2 = FactView(_("Production"), "6.2")
    v2.q_fact_keys = ["f"]  # Generated on the fly.
    v2.a_fact_keys = ["a", "p_1", "m_1", "n"] # Generated on the fly.

    fact_views = [v1, v2]
    unique_fact_keys = ["f"]
    required_fact_keys = ["f"]

    def fact_key_format_proxies(self):
        proxies = CardType.fact_key_format_proxies(self)
        proxies["a"] = "f"
        return proxies

    def is_fact_data_valid(self, fact_data):
        return CardType.is_fact_data_valid(self, fact_data)

    def fact_data(self, card):
        data = copy.copy(card.fact.data)
        # Recognition card.
        if card.fact_view == self.fact_views[0]:
            question, answer = self.q_a_from_cloze\
                (card.fact["f"], -1)
        # Production card.
        else:
            question, answer = self.q_a_from_cloze\
                (card.fact["f"], card.extra_data["index"])
            # Entire sentence clozed.
            if question.strip() == "[...]" and "m_1" in data:
                question = data["m_1"]
                data["m_1"] = ""
        data["f"] = question
        data["a"] = answer
        return data

    def create_sister_cards(self, fact):
        cards = [Card(self, fact, self.fact_views[0])]
        for match in cloze_re.finditer(fact["f"]):
            card = Card(self, fact, self.fact_views[1])
            card.extra_data["cloze"] = match.group(1)
            card.extra_data["index"] = len(cards) - 1
            cards.append(card)
        return cards

    def edit_fact(self, fact, new_fact_data):
        return self._edit_clozes(fact, new_fact_data,
            "f", self.fact_views[1])


class SentencePlugin(Plugin):

    name = _("Sentence")
    description = _("""A card type using sentences to study foreign languages.\n
Apart from simple recognition of the sentence, you can also add production cards using close deletion.\nE.g. if in the sentence field you write "La [casa:house] es [grande:big]", you'll get cards with questions like "La [house] es grande".""")
    components = [Sentence]
    supported_API_level = 3
